#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#    Thanks to https://github.com/tilleyd/cec2017-py, this script is modified or adapted from this repo 

import numpy as np

def bent_cigar(x):
    sm = 0.0
    for i in range(1, len(x)):
        sm += x[i]*x[i]
    sm *= 10e6
    return x[0]*x[0] + sm

def sum_diff_pow(x):
    sm = 0.0
    for i in range(0, len(x)):
        sm += (abs(x[i])) ** (i+1)
    return sm

def zakharov(x):
    sms = 0.0
    sm = 0.0
    for i in range(0, len(x)):
        sms += x[i]*x[i]
        # Note: the i+1 term is not in the CEC function definitions, but is
        # in the code and in any definition you find online
        sm += (i+1)*x[i]
    sm = 0.5 * sm
    sm = sm * sm
    return sms + sm + (sm * sm)

def rosenbrock(x):
    x = 0.02048 * x + 1.0
    sm = 0
    for i in range(0, len(x)-1):
        t1 = x[i]*x[i] - x[i+1]
        t1 = 100*t1*t1
        t2 = x[i] - 1
        t2 = t2*t2
        sm += t1 + t2
    return sm

def rastrigin(x):
    # Note: the 0.0512 shrinking is omitted in the problem definitions but is
    # present in the provided code
    x = 0.0512 * x
    tpi = 2.0 * np.pi
    sm = 0.0
    cs = np.cos(tpi*x)
    for i in range(0, len(x)):
        sm += x[i]*x[i] - 10*cs[i]
    return sm + 10*len(x)

def expanded_schaffers_f6(x):
    sm = 0.0
    for i in range(0, len(x)-1):
        t = x[i]*x[i] + x[i+1]*x[i+1]
        t1 = np.sin(np.sqrt(t))
        t1 = t1*t1 - 0.5
        t2 = 1 + 0.001*t
        t2 = t2*t2
        sm += 0.5 + t1/t2
    return sm

def lunacek_bi_rastrigin(x, shift=None, rotation=None):
    # a special case; we need the shift vector and rotation matrix
    nx = len(x)
    if shift is None:
        shift = np.zeros(nx)

    # calculate the coefficients
    mu0=2.5
    tmpx = np.zeros(nx)
    s = 1 - 1 / (2 * ((nx+20)**0.5) - 8.2)
    mu1 = -((mu0*mu0-1)/s)**0.5

    # shift and scale
    y = 0.1 * (x - shift)

    for i in range(0, nx):
        tmpx[i] = 2*y[i]
        if shift[i] < 0.0:
            tmpx[i] *= -1.0

    z = tmpx.copy()
    tmpx = tmpx + mu0

    t1=0.0
    t2=0.0
    for i in range(0, nx):
        t = tmpx[i]-mu0
        t1 += t*t
        t = tmpx[i]-mu1
        t2 += t*t
    t2 *= s
    t2 += nx

    y = z if rotation is None else np.matmul(rotation, z)

    t = 0.0
    y = np.cos(2.0*np.pi*y)
    for i in range(0, nx):
        t += y[i]

    r = t1 if t1 < t2 else t2
    return r + 10.0*(nx-t)

def non_cont_rastrigin(x, shift=None, rotation=None):
    # a special case; we need the shift vector and rotation matrix
    if shift is None:
        shift = np.zeros(x.shape)

    nx = len(x)
    sm = 0.0
    for i in range(0, nx):
        if abs(x[i]-shift[i]) > 0.5:
            x[i] = shift[i] + np.floor(2*(x[i]-shift[i])+0.5)/2

    z = 0.0512 * (x - shift)
    z = z if rotation is None else np.matmul(rotation, z)

    for i in range(0, nx):
        sm += (z[i]*z[i] - 10.0*np.cos(2.0*np.pi*z[i]) + 10.0)
    return sm

def levy(x):
    # Note: the function definitions state to scale by 5.12/100, but the code
    # doesn't do this, and the example graph in the definitions correspond to
    # the version without scaling
    # x = 0.0512 * x
    nx = len(x)
    w = 1.0 + 0.25*(x - 1.0)

    term1 = (np.sin(np.pi*w[0]))**2
    term3 = ((w[nx-1] - 1)**2) * (1 + ((np.sin(2*np.pi*w[nx-1]))**2))

    sm = 0.0

    for i in range(0, nx-1):
        wi = w[i]
        newv = ((wi-1)**2) * (1 + 10*((np.sin(np.pi*wi+1))**2))
        sm += newv

    return term1 + sm + term3

def modified_schwefel(x):
    nx = len(x)
    x = 10.0 * x # scale to search range
    sm = 0.0
    for i in range(0, nx):
        z = x[i] + 420.9687462275036
        if z < -500:
            zm = (abs(z) % 500) - 500
            t = z + 500
            t = t*t
            sm += zm * np.sin(np.sqrt(abs(zm))) - t / (10000*nx)
        elif z > 500:
            zm = 500 - (z % 500)
            t = z - 500
            t = t*t
            sm += zm * np.sin(np.sqrt(abs(zm))) - t / (10000*nx)
        else:
            sm += z * np.sin(np.sqrt(abs(z)))

    return 418.9829*nx - sm

def high_conditioned_elliptic(x):
    factor = 6 / (len(x) - 1)
    sm = 0.0
    for i in range(0, len(x)):
        sm += x[i]*x[i] * 10**(i*factor)
    return sm

def discus(x):
    sm = 1e+6*x[0]*x[0]
    for i in range(1, len(x)):
        sm += x[i]*x[i]
    return sm

def ackley(x):
    smsq = 0.0
    smcs = 0.0
    cs = np.cos((2*np.pi)*x)
    for i in range(0, len(x)):
        smsq += x[i]*x[i]
        smcs += cs[i]
    inx = 1/len(x)
    return -20*np.exp(-0.2*np.sqrt(inx*smsq)) - np.exp(inx*smcs) + 20 + np.e

def weierstrass(x):
    x = 0.005 * x
    k = np.arange(start=0, stop=21, step=1)
    ak = 0.5**k
    bk = np.pi * (3**k)
    sm = 0.0
    for i in range(0, len(x)):
        kcs = ak * np.cos(2*(x[i]+0.5)*bk)
        ksm = 0.0
        for j in range(0, 21):
            ksm += kcs[j]
        sm += ksm
    kcs = ak * np.cos(bk)
    ksm = 0.0
    for j in range(0, 21):
        ksm += kcs[j]
    return sm - len(x)*ksm

def griewank(x):
    x = 6.0 * x
    factor = 1/4000
    cs = np.cos(x / np.arange(start=1, stop=len(x)+1))
    sm = 0.0
    pd = 1.0
    for i in range(0, len(x)):
        sm += factor*x[i]*x[i]
        pd *= cs[i]
    return sm - pd + 1

def katsuura(x):
    x = 0.05 * x
    nx = len(x)
    pw = 10/(nx**1.2)
    prd = 1.0
    tj = 2**np.arange(start=1, stop=33, step=1)
    for i in range(0, nx):
        tjx = tj*x[i]
        t = np.abs(tjx - np.round(tjx)) / tj
        tsm = 0.0
        for j in range(0, 32):
            tsm += t[j]
        prd *= (1+ (i+1)*tsm)**pw
    df = 10/(nx*nx)
    return df*prd - df

def happy_cat(x):
    x = (0.05 * x) - 1
    nx = len(x)
    sm = 0.0
    smsq = 0.0
    for i in range(0, nx):
        sm += x[i]
        smsq += x[i]*x[i]
    return (abs(smsq - nx))**0.25 + (0.5*smsq + sm)/nx + 0.5

def h_g_bat(x):
    x = (0.05 * x) - 1
    nx = len(x)
    sm = 0.0
    smsq = 0.0
    for i in range(0, nx):
        sm += x[i]
        smsq += x[i]*x[i]
    return (abs(smsq*smsq - sm*sm))**0.5 + (0.5*smsq + sm)/nx + 0.5

def expanded_griewanks_plus_rosenbrock(x):
    x = (0.05 * x) + 1

    sm = 0.0
    for i in range(0, len(x)-1):
        tmp1 = x[i]*x[i]-x[i+1]
        tmp2 = x[i] - 1.0
        temp = 100*tmp1*tmp1 + tmp2*tmp2
        sm += (temp*temp)/4000.0 - np.cos(temp) + 1
        tmp1 = x[-1]*x[-1] - x[0]
        tmp2 = x[-1] - 1
        temp = 100.0*tmp1*tmp1 + tmp2*tmp2
        sm += (temp*temp)/4000.0 - np.cos(temp) + 1.0
    return sm

def schaffers_f7(x):
    nx = len(x)
    # Note: the function definitions state to scale by 0.5/100, but the code
    # doesn't do this, and the example graph in the definitions correspond to
    # the version without scaling
    # x = 0.005 * x
    sm = 0.0
    for i in range(0, nx-1):
        si = (x[i]*x[i] + x[i+1]*x[i+1])**0.5
        tmp = np.sin(50.0*(si**0.2))
        # Note: the original code has this error here (tmp shouldn't be squared)
        # that I'm keeping for consistency.
        sm += (si**0.5) * (tmp*tmp + 1)
    sm = (sm*sm) / (nx*nx - 2*nx + 1)
    return sm

all_functions = [
    bent_cigar,
    sum_diff_pow,
    zakharov,
    rosenbrock,
    rastrigin,
    expanded_schaffers_f6,
    lunacek_bi_rastrigin,
    non_cont_rastrigin,
    levy,
    modified_schwefel,
    high_conditioned_elliptic,
    discus,
    ackley,
    weierstrass,
    griewank,
    katsuura,
    happy_cat,
    h_g_bat,
    expanded_griewanks_plus_rosenbrock,
    schaffers_f7
]



import pickle
import warnings

import numpy as np

from neorl.rl.baselines.shared.vec_env.base_vec_env import VecEnvWrapper
from neorl.rl.baselines.shared.running_mean_std import RunningMeanStd


class VecNormalize(VecEnvWrapper):
    """
    A moving average, normalizing wrapper for vectorized environment.

    It is pickleable which will save moving averages and configuration parameters.
    The wrapped environment `venv` is not saved, and must be restored manually with
    `set_venv` after being unpickled.

    :param venv: (VecEnv) the vectorized environment to wrap
    :param training: (bool) Whether to update or not the moving average
    :param norm_obs: (bool) Whether to normalize observation or not (default: True)
    :param norm_reward: (bool) Whether to normalize rewards or not (default: True)
    :param clip_obs: (float) Max absolute value for observation
    :param clip_reward: (float) Max value absolute for discounted reward
    :param gamma: (float) discount factor
    :param epsilon: (float) To avoid division by zero
    """

    def __init__(self, venv, training=True, norm_obs=True, norm_reward=True,
                 clip_obs=10., clip_reward=10., gamma=0.99, epsilon=1e-8):
        VecEnvWrapper.__init__(self, venv)
        self.obs_rms = RunningMeanStd(shape=self.observation_space.shape)
        self.ret_rms = RunningMeanStd(shape=())
        self.clip_obs = clip_obs
        self.clip_reward = clip_reward
        # Returns: discounted rewards
        self.ret = np.zeros(self.num_envs)
        self.gamma = gamma
        self.epsilon = epsilon
        self.training = training
        self.norm_obs = norm_obs
        self.norm_reward = norm_reward
        self.old_obs = None
        self.old_rews = None

    def __getstate__(self):
        """
        Gets state for pickling.

        Excludes self.venv, as in general VecEnv's may not be pickleable."""
        state = self.__dict__.copy()
        # these attributes are not pickleable
        del state['venv']
        del state['class_attributes']
        # these attributes depend on the above and so we would prefer not to pickle
        del state['ret']
        return state

    def __setstate__(self, state):
        """
        Restores pickled state.

        User must call set_venv() after unpickling before using.

        :param state: (dict)"""
        self.__dict__.update(state)
        assert 'venv' not in state
        self.venv = None

    def set_venv(self, venv):
        """
        Sets the vector environment to wrap to venv.

        Also sets attributes derived from this such as `num_env`.

        :param venv: (VecEnv)
        """
        if self.venv is not None:
            raise ValueError("Trying to set venv of already initialized VecNormalize wrapper.")
        VecEnvWrapper.__init__(self, venv)
        if self.obs_rms.mean.shape != self.observation_space.shape:
            raise ValueError("venv is incompatible with current statistics.")
        self.ret = np.zeros(self.num_envs)

    def step_wait(self):
        """
        Apply sequence of actions to sequence of environments
        actions -> (observations, rewards, news)

        where 'news' is a boolean vector indicating whether each element is new.
        """
        obs, rews, news, infos = self.venv.step_wait()
        self.old_obs = obs
        self.old_rews = rews

        if self.training:
            self.obs_rms.update(obs)
        obs = self.normalize_obs(obs)

        if self.training:
            self._update_reward(rews)
        rews = self.normalize_reward(rews)

        self.ret[news] = 0
        return obs, rews, news, infos

    def _update_reward(self, reward: np.ndarray) -> None:
        """Update reward normalization statistics."""
        self.ret = self.ret * self.gamma + reward
        self.ret_rms.update(self.ret)

    def normalize_obs(self, obs: np.ndarray) -> np.ndarray:
        """
        Normalize observations using this VecNormalize's observations statistics.
        Calling this method does not update statistics.
        """
        if self.norm_obs:
            obs = np.clip((obs - self.obs_rms.mean) / np.sqrt(self.obs_rms.var + self.epsilon),
                          -self.clip_obs,
                          self.clip_obs)
        return obs

    def normalize_reward(self, reward: np.ndarray) -> np.ndarray:
        """
        Normalize rewards using this VecNormalize's rewards statistics.
        Calling this method does not update statistics.
        """
        if self.norm_reward:
            reward = np.clip(reward / np.sqrt(self.ret_rms.var + self.epsilon),
                           -self.clip_reward, self.clip_reward)
        return reward

    def get_original_obs(self) -> np.ndarray:
        """
        Returns an unnormalized version of the observations from the most recent
        step or reset.
        """
        return self.old_obs.copy()

    def get_original_reward(self) -> np.ndarray:
        """
        Returns an unnormalized version of the rewards from the most recent step.
        """
        return self.old_rews.copy()

    def reset(self):
        """
        Reset all environments
        """
        obs = self.venv.reset()
        self.old_obs = obs
        self.ret = np.zeros(self.num_envs)
        if self.training:
            self._update_reward(self.ret)
        return self.normalize_obs(obs)

    @staticmethod
    def load(load_path, venv):
        """
        Loads a saved VecNormalize object.

        :param load_path: the path to load from.
        :param venv: the VecEnv to wrap.
        :return: (VecNormalize)
        """
        with open(load_path, "rb") as file_handler:
            vec_normalize = pickle.load(file_handler)
        vec_normalize.set_venv(venv)
        return vec_normalize

    def save(self, save_path):
        with open(save_path, "wb") as file_handler:
            pickle.dump(self, file_handler)

    def save_running_average(self, path):
        """
        :param path: (str) path to log dir

        .. deprecated:: 2.9.0
            This function will be removed in a future version
        """
        warnings.warn("Usage of `save_running_average` is deprecated. Please "
                      "use `save` or pickle instead.", DeprecationWarning)
        for rms, name in zip([self.obs_rms, self.ret_rms], ['obs_rms', 'ret_rms']):
            with open("{}/{}.pkl".format(path, name), 'wb') as file_handler:
                pickle.dump(rms, file_handler)

    def load_running_average(self, path):
        """
        :param path: (str) path to log dir

        .. deprecated:: 2.9.0
            This function will be removed in a future version
        """
        warnings.warn("Usage of `load_running_average` is deprecated. Please "
                      "use `load` or pickle instead.", DeprecationWarning)
        for name in ['obs_rms', 'ret_rms']:
            with open("{}/{}.pkl".format(path, name), 'rb') as file_handler:
                setattr(self, name, pickle.load(file_handler))



import os
from abc import ABC
import warnings
import typing
from typing import Union, List, Dict, Any, Optional

import gym
import numpy as np

from neorl.rl.baselines.shared.vec_env import VecEnv, sync_envs_normalization, DummyVecEnv
from neorl.rl.baselines.shared.evaluation import evaluate_policy
from neorl.rl.baselines.shared import logger

if typing.TYPE_CHECKING:
    from neorl.rl.baselines.shared.base_class import BaseRLModel  # pytype: disable=pyi-error


class BaseCallback(ABC):
    """
    Base class for callback.

    :param verbose: (int)
    """
    def __init__(self, verbose: int = 0):
        super(BaseCallback, self).__init__()
        # The RL model
        self.model = None  # type: Optional[BaseRLModel]
        # An alias for self.model.get_env(), the environment used for training
        self.training_env = None  # type: Union[gym.Env, VecEnv, None]
        # Number of time the callback was called
        self.n_calls = 0  # type: int
        # n_envs * n times env.step() was called
        self.num_timesteps = 0  # type: int
        self.verbose = verbose
        self.locals = None  # type: Optional[Dict[str, Any]]
        self.globals = None  # type: Optional[Dict[str, Any]]
        self.logger = None  # type: Optional[logger.Logger]
        # Sometimes, for event callback, it is useful
        # to have access to the parent object
        self.parent = None  # type: Optional[BaseCallback]

    # Type hint as string to avoid circular import
    def init_callback(self, model: 'BaseRLModel') -> None:
        """
        Initialize the callback by saving references to the
        RL model and the training environment for convenience.
        """
        self.model = model
        self.training_env = model.get_env()
        self.logger = logger.Logger.CURRENT
        self._init_callback()

    def update_locals(self, locals_: Dict[str, Any]) -> None:
        """
        Updates the local variables of the training process

        For reference to which variables are accessible,
        check each individual algorithm's documentation
        :param `locals_`: (Dict[str, Any]) current local variables
        """
        self.locals.update(locals_)

    def _init_callback(self) -> None:
        pass

    def on_training_start(self, locals_: Dict[str, Any], globals_: Dict[str, Any]) -> None:
        # Those are reference and will be updated automatically
        self.locals = locals_
        self.globals = globals_
        self._on_training_start()

    def _on_training_start(self) -> None:
        pass

    def on_rollout_start(self) -> None:
        self._on_rollout_start()

    def _on_rollout_start(self) -> None:
        pass

    def _on_step(self) -> bool:
        """
        :return: (bool) If the callback returns False, training is aborted early.
        """
        return True

    def on_step(self) -> bool:
        """
        This method will be called by the model after each call to `env.step()`.

        For child callback (of an `EventCallback`), this will be called
        when the event is triggered.

        :return: (bool) If the callback returns False, training is aborted early.
        """
        self.n_calls += 1
        self.num_timesteps = self.model.num_timesteps

        return self._on_step()

    def on_training_end(self) -> None:
        self._on_training_end()

    def _on_training_end(self) -> None:
        pass

    def on_rollout_end(self) -> None:
        self._on_rollout_end()

    def _on_rollout_end(self) -> None:
        pass


class EventCallback(BaseCallback):
    """
    Base class for triggering callback on event.

    :param callback: (Optional[BaseCallback]) Callback that will be called
        when an event is triggered.
    :param verbose: (int)
    """
    def __init__(self, callback: Optional[BaseCallback] = None, verbose: int = 0):
        super(EventCallback, self).__init__(verbose=verbose)
        self.callback = callback
        # Give access to the parent
        if callback is not None:
            self.callback.parent = self

    def init_callback(self, model: 'BaseRLModel') -> None:
        super(EventCallback, self).init_callback(model)
        if self.callback is not None:
            self.callback.init_callback(self.model)

    def _on_training_start(self) -> None:
        if self.callback is not None:
            self.callback.on_training_start(self.locals, self.globals)

    def _on_event(self) -> bool:
        if self.callback is not None:
            return self.callback.on_step()
        return True

    def _on_step(self) -> bool:
        return True


class CallbackList(BaseCallback):
    """
    Class for chaining callbacks.

    :param callbacks: (List[BaseCallback]) A list of callbacks that will be called
        sequentially.
    """
    def __init__(self, callbacks: List[BaseCallback]):
        super(CallbackList, self).__init__()
        assert isinstance(callbacks, list)
        self.callbacks = callbacks

    def _init_callback(self) -> None:
        for callback in self.callbacks:
            callback.init_callback(self.model)

    def _on_training_start(self) -> None:
        for callback in self.callbacks:
            callback.on_training_start(self.locals, self.globals)

    def _on_rollout_start(self) -> None:
        for callback in self.callbacks:
            callback.on_rollout_start()

    def _on_step(self) -> bool:
        continue_training = True
        for callback in self.callbacks:
            # Return False (stop training) if at least one callback returns False
            continue_training = callback.on_step() and continue_training
        return continue_training

    def _on_rollout_end(self) -> None:
        for callback in self.callbacks:
            callback.on_rollout_end()

    def _on_training_end(self) -> None:
        for callback in self.callbacks:
            callback.on_training_end()


class CheckpointCallback(BaseCallback):
    """
    Callback for saving a model every `save_freq` steps

    :param save_freq: (int)
    :param save_path: (str) Path to the folder where the model will be saved.
    :param name_prefix: (str) Common prefix to the saved models
    """
    def __init__(self, save_freq: int, save_path: str, name_prefix='rl_model', verbose=0):
        super(CheckpointCallback, self).__init__(verbose)
        self.save_freq = save_freq
        self.save_path = save_path
        self.name_prefix = name_prefix

    def _init_callback(self) -> None:
        # Create folder if needed
        if self.save_path is not None:
            os.makedirs(self.save_path, exist_ok=True)

    def _on_step(self) -> bool:
        if self.n_calls % self.save_freq == 0:
            path = os.path.join(self.save_path, '{}_{}_steps'.format(self.name_prefix, self.num_timesteps))
            self.model.save(path)
            if self.verbose > 1:
                print("Saving model checkpoint to {}".format(path))
        return True


class ConvertCallback(BaseCallback):
    """
    Convert functional callback (old-style) to object.

    :param callback: (Callable)
    :param verbose: (int)
    """
    def __init__(self, callback, verbose=0):
        super(ConvertCallback, self).__init__(verbose)
        self.callback = callback

    def _on_step(self) -> bool:
        if self.callback is not None:
            return self.callback(self.locals, self.globals)
        return True


class EvalCallback(EventCallback):
    """
    Callback for evaluating an agent.

    :param eval_env: (Union[gym.Env, VecEnv]) The environment used for initialization
    :param callback_on_new_best: (Optional[BaseCallback]) Callback to trigger
        when there is a new best model according to the `mean_reward`
    :param n_eval_episodes: (int) The number of episodes to test the agent
    :param eval_freq: (int) Evaluate the agent every eval_freq call of the callback.
    :param log_path: (str) Path to a folder where the evaluations (`evaluations.npz`)
        will be saved. It will be updated at each evaluation.
    :param best_model_save_path: (str) Path to a folder where the best model
        according to performance on the eval env will be saved.
    :param deterministic: (bool) Whether the evaluation should
        use a stochastic or deterministic actions.
    :param render: (bool) Whether to render or not the environment during evaluation
    :param verbose: (int)
    """
    def __init__(self, eval_env: Union[gym.Env, VecEnv],
                 callback_on_new_best: Optional[BaseCallback] = None,
                 n_eval_episodes: int = 5,
                 eval_freq: int = 10000,
                 log_path: str = None,
                 best_model_save_path: str = None,
                 deterministic: bool = True,
                 render: bool = False,
                 verbose: int = 1):
        super(EvalCallback, self).__init__(callback_on_new_best, verbose=verbose)
        self.n_eval_episodes = n_eval_episodes
        self.eval_freq = eval_freq
        self.best_mean_reward = -np.inf
        self.last_mean_reward = -np.inf
        self.deterministic = deterministic
        self.render = render

        # Convert to VecEnv for consistency
        if not isinstance(eval_env, VecEnv):
            eval_env = DummyVecEnv([lambda: eval_env])

        assert eval_env.num_envs == 1, "You must pass only one environment for evaluation"

        self.eval_env = eval_env
        self.best_model_save_path = best_model_save_path
        # Logs will be written in `evaluations.npz`
        if log_path is not None:
            log_path = os.path.join(log_path, 'evaluations')
        self.log_path = log_path
        self.evaluations_results = []
        self.evaluations_timesteps = []
        self.evaluations_length = []

    def _init_callback(self):
        # Does not work in some corner cases, where the wrapper is not the same
        if not type(self.training_env) is type(self.eval_env):
            warnings.warn("Training and eval env are not of the same type"
                          "{} != {}".format(self.training_env, self.eval_env))

        # Create folders if needed
        if self.best_model_save_path is not None:
            os.makedirs(self.best_model_save_path, exist_ok=True)
        if self.log_path is not None:
            os.makedirs(os.path.dirname(self.log_path), exist_ok=True)

    def _on_step(self) -> bool:

        if self.eval_freq > 0 and self.n_calls % self.eval_freq == 0:
            # Sync training and eval env if there is VecNormalize
            sync_envs_normalization(self.training_env, self.eval_env)

            episode_rewards, episode_lengths = evaluate_policy(self.model, self.eval_env,
                                                               n_eval_episodes=self.n_eval_episodes,
                                                               render=self.render,
                                                               deterministic=self.deterministic,
                                                               return_episode_rewards=True)

            if self.log_path is not None:
                self.evaluations_timesteps.append(self.num_timesteps)
                self.evaluations_results.append(episode_rewards)
                self.evaluations_length.append(episode_lengths)
                np.savez(self.log_path, timesteps=self.evaluations_timesteps,
                         results=self.evaluations_results, ep_lengths=self.evaluations_length)

            mean_reward, std_reward = np.mean(episode_rewards), np.std(episode_rewards)
            mean_ep_length, std_ep_length = np.mean(episode_lengths), np.std(episode_lengths)
            # Keep track of the last evaluation, useful for classes that derive from this callback
            self.last_mean_reward = mean_reward

            if self.verbose > 0:
                print("Eval num_timesteps={}, "
                      "episode_reward={:.2f} +/- {:.2f}".format(self.num_timesteps, mean_reward, std_reward))
                print("Episode length: {:.2f} +/- {:.2f}".format(mean_ep_length, std_ep_length))

            if mean_reward > self.best_mean_reward:
                if self.verbose > 0:
                    print("New best mean reward!")
                if self.best_model_save_path is not None:
                    self.model.save(os.path.join(self.best_model_save_path, 'best_model'))
                self.best_mean_reward = mean_reward
                # Trigger callback if needed
                if self.callback is not None:
                    return self._on_event()

        return True


class StopTrainingOnRewardThreshold(BaseCallback):
    """
    Stop the training once a threshold in episodic reward
    has been reached (i.e. when the model is good enough).

    It must be used with the `EvalCallback`.

    :param reward_threshold: (float)  Minimum expected reward per episode
        to stop training.
    :param verbose: (int)
    """
    def __init__(self, reward_threshold: float, verbose: int = 0):
        super(StopTrainingOnRewardThreshold, self).__init__(verbose=verbose)
        self.reward_threshold = reward_threshold

    def _on_step(self) -> bool:
        assert self.parent is not None, ("`StopTrainingOnRewardThreshold` callback must be used "
                                         "with an `EvalCallback`")
        # Convert np.bool to bool, otherwise callback.on_step() is False won't work
        continue_training = bool(self.parent.best_mean_reward < self.reward_threshold)
        if self.verbose > 0 and not continue_training:
            print("Stopping training because the mean reward {:.2f} "
                  " is above the threshold {}".format(self.parent.best_mean_reward, self.reward_threshold))
        return continue_training


class EveryNTimesteps(EventCallback):
    """
    Trigger a callback every `n_steps` timesteps

    :param n_steps: (int) Number of timesteps between two trigger.
    :param callback: (BaseCallback) Callback that will be called
        when the event is triggered.
    """
    def __init__(self, n_steps: int, callback: BaseCallback):
        super(EveryNTimesteps, self).__init__(callback)
        self.n_steps = n_steps
        self.last_time_trigger = 0

    def _on_step(self) -> bool:
        if (self.num_timesteps - self.last_time_trigger) >= self.n_steps:
            self.last_time_trigger = self.num_timesteps
            return self._on_event()
        return True



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.


"""
This class parses the master input file provided by the user
""" 
import multiprocessing, os, subprocess
import numpy as np
import json
from ast import literal_eval

def str_to_bool(s):
    if s == 'True':
         return True
    elif s == 'False':
         return False
    else:
         raise ValueError # evil ValueError that doesn't tell you what the wrong value was

class InputParser():
    def __init__(self, input_file_path):
        
        #infer the casename from filename
        try:
            self.MainCaseName=input_file_path.split(".")[0]
            self.log_dir=self.MainCaseName+'_log/'
        except:
            self.log_dir=input_file_path+'_log/'
        #-----------------
        # method flags initialize
        #-----------------
        self.dqn_flag=False
        self.ga_flag=False
        self.a2c_flag=False
        self.acer_flag=False
        self.ppo_flag=False
        self.sa_flag=False
        self.tune_flag=False
        # add new block flag
        
        #-----------------
        #input blocks 
        #-----------------
        self.gen_block={}
        self.dqn_block={}
        self.ppo_block={}
        self.a2c_block={}
        self.acer_block={}
        self.ga_block={}
        self.sa_block={}
        self.tune_block={}
        # add new block 
        
        #------------------------------------------------------------------
        #Read the lines of the input file into the attribute input_file
        #Remove first line, empty lines and "\n" from every line
        #------------------------------------------------------------------
        #old way of parsing the input
#        with open(input_file_path) as input_file_text:
#            self.input_file = [parameter.replace("\n","").strip() for parameter in input_file_text.readlines() if not parameter.isspace() or parameter[0] != "#"]
        
        #-- Paul parser
        self.input_file = []
        input_file_text = open(input_file_path,'r', encoding='utf-8')
        while True:
            line1 = input_file_text.readline()
            if not line1:break
            else:
                if 'env_data' in line1:
                    line1 = line1[:line1.rfind("#")]
                    temp_data = line1.strip()
                    pairing1 = [elem for elem,x in enumerate(temp_data) if x == '{'] 
                    pairing2 = [elem for elem,x in enumerate(temp_data) if x == '}']
                    while len(pairing1) != len(pairing2):
                        line1 = input_file_text.readline()
                        line1 = line1[:line1.rfind("#")]
                        temp_data += line1.strip()
                        pairing1 = [elem for elem,x in enumerate(temp_data) if x == '{'] 
                        pairing2 = [elem for elem,x in enumerate(temp_data) if x == '}']
                    self.input_file.append(temp_data)
                else:
                    self.input_file.append(line1.strip())
        input_file_text.close()
        #print(self.input_file)
        # check first character in each line, if invalid, raise an error 
        for item in self.input_file:
            if item != '':
                if not item[0].isalpha() and item[0] not in ['#', '{']:
                    print('*******************************')
                    print('illegal character is identified')
                    print('*******************************')
                    print(item)
                    raise Exception('Only alphapet or # is allowed as first character per line')
                
        # remove empty strings 
        self.input_file=[item for item in self.input_file if item != '']
        # remove comments  #----
        self.input_file=[item for item in self.input_file if item[0] != '#']
        
        #------------------------------
        # check GENERAL card 
        #------------------------------
        try:
            gen_test = self.input_file[self.input_file.index("READ GENERAL") + 1 : self.input_file.index("END GENERAL")]
            print ('--debug: General block is identified')
        except:
            raise('ERR: The general block is required but not found in the input --> READ GENERAL ... END GENERAL is missing')
        
        
    def parse_card(self, card):
        """
        This is a function that gets the Block/Card name, and parse it in self.input_file!
        Returns: a parsed dictionary for the block GENERAL, PPO, etc.
        """
        
        block = self.input_file[self.input_file.index("READ {}".format(card)) + 1 : self.input_file.index("END {}".format(card))]
        block = [item for item in block if item[0] != "#"]   # dobule check if there is any comments
        block = [item.split("=") for item in block]
        block = [[element.strip() for element in item] for item in block] #strip all white spaces
        
        card_dict={}
        for item in block:
            #----------------------------------------
            # convert string dict to a real dict
            if 'env_data' in item:
                #print(item[1])
                try:
                    item[1] = json.loads(item[1])
                except:
                    try:
                        item[1] = literal_eval(item[1])
                    except:
                        raise ValueError ('--error: env_data dict cannot be converted to a python dict, the parser failed')                   
            
                #print(item)
            #----------------------------------------
            card_dict[item[0]] = item[1]  
        
        return card_dict
        
    def blocks (self):

        """
        This function moves block by block and uses parse_card to parse each block
        Returns: a parsed dictionary for each block and flag for the activated blocks
        """
        #Setting attributes for each block, move one by one, check if they are available
        self.gen_block=self.parse_card("GENERAL")
        flags=[]
        #-----------  
        if ("READ TUNE" in self.input_file and "END TUNE" in self.input_file):
            print ('--debug: TUNE block is identified')
            self.tune_block=self.parse_card("TUNE")
            self.tune_flag=True; 
            if 'method' not in self.tune_block.keys():
                raise Exception ('--error: parameter ({}) in card {} is required for neorl but it is not given in the input'.format('method', 'TUNE'))
            #self.tune_block['log_dir'] = self.log_dir  #add log_dir to tuneblock
        #-----------  
        if ("READ DQN" in self.input_file and "END DQN" in self.input_file):
            print ('--debug: DQN block is identified')
            self.dqn_block=self.parse_card("DQN")
            self.dqn_flag=True; flags.append(1)
        #-----------  
        if ("READ PPO" in self.input_file and "END PPO" in self.input_file):
            print ('--debug: PPO block is identified')
            self.ppo_block=self.parse_card("PPO")
            self.ppo_flag=True; flags.append(1)
        #-----------    
        if ("READ A2C" in self.input_file and "END A2C" in self.input_file):
            print ('--debug: A2C block is identified')
            self.a2c_block=self.parse_card("A2C")
            self.a2c_flag=True; flags.append(1)
        #-----------    
        if ("READ ACER" in self.input_file and "END ACER" in self.input_file):
            print ('--debug: ACER block is identified')
            self.acer_block=self.parse_card("ACER")
            self.acer_flag=True; flags.append(1)
        #-----------  
        if ("READ GA" in self.input_file and "END GA" in self.input_file):
            print ('--debug: GA block is identified')
            self.ga_block=self.parse_card("GA")
            self.ga_flag=True; flags.append(1)
        #-----------  
        if ("READ SA" in self.input_file and "END SA" in self.input_file):
            print ('--debug: SA block is identified')
            self.sa_block=self.parse_card("SA")
            self.sa_flag=True; flags.append(1)
        #-----------    
        if self.tune_flag:
            if len(flags) > 1:
                raise Exception ('only one algorithm block can be used with TUNE, user provided {} blocks'.format(len(flags)))
            elif len(flags) == 0:
                raise Exception ('user defined TUNE block but did not use any algorithm block with it, e.g. DQN, GA, SA, etc.')
        return self.log_dir

class InputChecker(InputParser):
    
    """
    This class checks the input for any errors/typos and then overwrites the default inputs by user ones
    Inputs:
        master_parser: are the parsed paramters from class InputParser
        master_paramdict: are the default values given from the class ParamList.py
    """ 
    def __init__(self, master_parser, master_paramdict, log_dir):
        self.paramdict=master_paramdict
        self.parser=master_parser
        self.log_dir=log_dir
        #-----------------------------------------------------------------
        #Initialize all dictionaries to the default ones in ParamList.py
        #-----------------------------------------------------------------
        self.gen_dict=self.paramdict.gen_dict
        self.dqn_dict=self.paramdict.dqn_dict
        self.a2c_dict=self.paramdict.a2c_dict
        self.acer_dict=self.paramdict.acer_dict
        self.ppo_dict=self.paramdict.ppo_dict
        self.ga_dict=self.paramdict.ga_dict
        self.sa_dict=self.paramdict.sa_dict
                    
    def check_input (self, parser, paramdict, card):
        """
        This function loops through any data list and check if data structure and types are correct
        Inputs: 
            parser: is a parsed dict from the user for any block 
            paramdict: is a default dict by neorl for any block
            card: is the block/card name
        
        Returns: this function does not return, but overwrites the default values in self.gen_dict, self.dqn_dict, ...
        """        
        for item in parser:
            if item not in paramdict:
                print('--error: {} is NOT found in neorl input variable names'.format(item))
                raise(ValueError)
                            
            try: 
                if paramdict[item][2] == "str":
                    parser[item] = str(parser[item]).strip()
                elif paramdict[item][2] == "int":
                    parser[item] = int(float(parser[item]))
                elif paramdict[item][2] == "float":
                    parser[item] = float(parser[item])
                elif paramdict[item][2] == "bool":
                    parser[item] = str_to_bool(parser[item])
                elif paramdict[item][2] == "strvec":
                    parser[item] = [str(element.strip()) for element in parser[item].split(",")]
                elif paramdict[item][2] == "vec":
                    parser[item] = np.array([float(element.strip()) for element in parser[item].split(",")])
            except:
                print('--error: the data structure for parameter {} in card {} must be {}, but something else is used'.format(item, card, paramdict[item][2]))
                raise(ValueError)
                
        for item in paramdict:
            if paramdict[item][1] == "r" and item not in parser.keys():
                raise Exception ('--error: parameter {} in card {} is required for neorl but it is not given in the input'.format(item, card))
                
            if paramdict[item][1] == "o" and item not in parser.keys():
                if item not in ['flag']:
                    print ('--warning: parameter {} in card {} is missed, Default is used ---> {}'.format(item,card, paramdict[item][0]))
            
            if paramdict[item][1] == "rs" and item not in parser.keys():
                
                if item == 'model_load_path' and parser['mode'][0] in ['test','continue']: #testing or continue learning without path to pre-trained model
                    raise Exception('--error: the user selected test mode in card {}, but no path to the pre-trained model is provided via model_load_path'.format(card))

                if item == 'time_steps' and parser['mode'][0] in ['train','continue']: #testing or continue learning without path to pre-trained model
                    raise Exception('--error: the user selected train or continue mode in card {}, but number of time_steps is not defined'.format(card))
                    
        # check the test 
        if card in ['DQN', 'PPO', 'A2C', 'ACER']:
            if parser['mode'] in ['train']:
                except_var=[i for i in ['model_load_path', 'n_eval_episodes', 'render'] if i in parser.keys()]
                if len(except_var) > 0:
                    raise Exception('--error: the following variables {} in card {} are NOT allowed for TRAINING mode'.format(except_var, card))
            
            if parser['mode'] in ['continue']:
                except_var=[i for i in ['n_eval_episodes', 'render'] if i in parser.keys()]
                if len(except_var) > 0:
                    raise Exception('--error: the following variables {} in card {} are NOT allowed for CONTINUAL mode'.format(except_var, card))

            if parser['mode'] in ['test']:
                test_allowed=['casename', 'mode', 'model_load_path', 'n_eval_episodes', 'render']
                except_var=[i for i in parser.keys() if i not in test_allowed]
                if len(except_var) > 0:
                    raise Exception('--error: the following variables {} in card {} are ONLY allowed for TEST mode'.format(test_allowed, card))
                   
        #check conditions in genral card
        if card in ['GENERAL']:
            if "xsize_plot" not in parser.keys():
                parser['xsize_plot'] = parser['xsize']
                print ('--warning: xsize_plot is set to equal to xsize ---> {}'.format(parser['xsize_plot']))
            
#            if "exepath" in parser.keys():
#                
#                print ('--debug: checking the exepath')
#                if os.path.isdir(parser['exepath']):
#                    raise Exception ('--error: the user provided a path for directory not to exefile --> {} --> not complete'.format(parser['exepath']))
#                execheck=os.system('which {}'.format(parser['exepath']))
#                if os.path.exists(parser['exepath']):
#                    print('--debug: User provided absolute directory and the binary file reported in {} exists'.format(parser['exepath']))
#                elif (execheck==0):
#                    exeinfer=subprocess.check_output(['which', str(parser['exepath'])])
#                    parser['exepath']=exeinfer.decode('utf-8').strip()
#                    print('--debug: neorl tried to infer the exepath via which and found {}'.format(parser['exepath']))
#                else:
#                    raise Exception ('--error: The binary file reported in {} cannot be found'.format(parser['exepath']))
#                    
            #------------------------------------
            #check availability of GYM env
            #------------------------------------
#            try: 
#                gym.make(parser['env'])
#                print('--debug: {} env is pre-registered in GYM'.format(parser['env']))
#            except:
#                print('--warning: {} env is NOT pre-registered in GYM'.format(parser['env']))
#                try:
#                    print('--warning: Attempting to register {} ...'.format(parser['env']))
#                except:
#                    raise Exception ('The enviroment failed to be registred in GYM, check the entry point to the env, or any syntax problems')
                
                    
    def setup_input(self):
        
        # check the strucutre and syntax first, then overwrite the default dictionary in paramdict.
        self.methods=[]
        self.used_cores=0
        self.check_input(self.parser.gen_block,self.gen_dict, 'GENERAL')
        maxcore_flag=False
        for item in self.parser.gen_block:
            self.gen_dict[item][0] = self.parser.gen_block[item]
            # General checks
            
            #----------------------------
            # check maxcores parameter
            #----------------------------
            
            if item == 'maxcores':
                print ('--debug: maxcores is given by user as {}'.format(self.parser.gen_block[item]))
                maxcore_flag=True
                self.max_cores=self.parser.gen_block[item]
                if self.parser.gen_block[item] <= 0:
                    self.parser.gen_block[item]=multiprocessing.cpu_count()
                    self.max_cores=self.parser.gen_block[item]
                    print ('--debug: user requested inference of maxcores for the machine which is {}'.format(self.max_cores))
        
        self.gen_dict['log_dir']=self.log_dir  # append the main log_dir as part of the gen_dict
        if not maxcore_flag:
            print('--warning: no limit on maxcores is provided by the user, so all specificed cores in the input will be used')
            
        #----------------------------
                
        if self.parser.dqn_flag:
            
            self.check_input(self.parser.dqn_block,self.dqn_dict, 'DQN')
            self.dqn_dict['flag'][0] = True
            for item in self.parser.dqn_block:
                self.dqn_dict[item][0] = self.parser.dqn_block[item]
            
            self.methods.append(self.dqn_dict['casename'][0])
            self.used_cores += self.dqn_dict["ncores"][0]
            

        if self.parser.ppo_flag:
            self.check_input(self.parser.ppo_block,self.ppo_dict, 'PPO')
            self.ppo_dict['flag'][0] = True
            for item in self.parser.ppo_block:
                self.ppo_dict[item][0] = self.parser.ppo_block[item]
            
            # adjust number of steps for parallel
            if self.ppo_dict["ncores"][0] > 1:
                if self.ppo_dict["check_freq"][0] % self.ppo_dict["ncores"][0] != 0:
                    mod=self.ppo_dict["check_freq"][0] % self.ppo_dict["ncores"][0]
                    print('warning: the check_freq parameter {} for ppo is not multiple of number of cores {}'.format(self.ppo_dict["check_freq"][0], self.ppo_dict["ncores"][0]))
                    self.ppo_dict["check_freq"][0] = (self.ppo_dict["check_freq"][0] + self.ppo_dict["ncores"][0] - mod)
                    assert (self.ppo_dict["check_freq"][0] % self.ppo_dict["ncores"][0]) == 0
                    print('warning: the check_freq parameter is adjusted to {} for ppo'.format(self.ppo_dict["check_freq"][0]))
                    
            
            self.methods.append(self.ppo_dict['casename'][0])
            self.used_cores += self.ppo_dict["ncores"][0]
            
        if self.parser.a2c_flag:
            self.check_input(self.parser.a2c_block,self.a2c_dict, 'A2C')
            self.a2c_dict['flag'][0] = True
            for item in self.parser.a2c_block:
                self.a2c_dict[item][0] = self.parser.a2c_block[item]
                            
            # adjust number of steps for parallel
            if self.a2c_dict["ncores"][0] > 1:
                if self.a2c_dict["check_freq"][0] % self.a2c_dict["ncores"][0] != 0:
                    mod=self.a2c_dict["check_freq"][0] % self.a2c_dict["ncores"][0]
                    print('warning: the check_freq parameter {} for a2c is not multiple of number of cores {}'.format(self.a2c_dict["check_freq"][0], self.a2c_dict["ncores"][0]))
                    self.a2c_dict["check_freq"][0] = (self.a2c_dict["check_freq"][0] + self.a2c_dict["ncores"][0] - mod)
                    assert (self.a2c_dict["check_freq"][0] % self.a2c_dict["ncores"][0]) == 0
                    print('warning: the check_freq parameter is adjusted to {} for a2c'.format(self.a2c_dict["check_freq"][0]))
                    
            self.methods.append(self.a2c_dict['casename'][0])
            self.used_cores += self.a2c_dict["ncores"][0]

        if self.parser.acer_flag:
            self.check_input(self.parser.acer_block,self.acer_dict, 'ACER')
            self.acer_dict['flag'][0] = True
            for item in self.parser.acer_block:
                self.acer_dict[item][0] = self.parser.acer_block[item]
                            
            # adjust number of steps for parallel
            if self.acer_dict["ncores"][0] > 1:
                if self.acer_dict["check_freq"][0] % self.acer_dict["ncores"][0] != 0:
                    mod=self.acer_dict["check_freq"][0] % self.acer_dict["ncores"][0]
                    print('warning: the check_freq parameter {} for acer is not multiple of number of cores {}'.format(self.acer_dict["check_freq"][0], self.acer_dict["ncores"][0]))
                    self.acer_dict["check_freq"][0] = (self.acer_dict["check_freq"][0] + self.acer_dict["ncores"][0] - mod)
                    assert (self.acer_dict["check_freq"][0] % self.acer_dict["ncores"][0]) == 0
                    print('warning: the check_freq parameter is adjusted to {} for acer'.format(self.acer_dict["check_freq"][0]))
                    
            self.methods.append(self.acer_dict['casename'][0])
            self.used_cores += self.acer_dict["ncores"][0]
            
        if self.parser.ga_flag:
            self.check_input(self.parser.ga_block,self.ga_dict, 'GA')
            self.ga_dict['flag'][0] = True
            for item in self.parser.ga_block:
                self.ga_dict[item][0] = self.parser.ga_block[item]
            
            self.methods.append(self.ga_dict['casename'][0])
            self.used_cores += self.ga_dict["ncores"][0]
                        
        if self.parser.sa_flag:
            self.check_input(self.parser.sa_block,self.sa_dict, 'SA')
            self.sa_dict['flag'][0] = True
            for item in self.parser.sa_block:
                self.sa_dict[item][0] = self.parser.sa_block[item]
            
            self.methods.append(self.sa_dict['casename'][0])
            self.used_cores += self.sa_dict["ncores"][0]            
        
        if maxcore_flag:
            assert self.used_cores <= self.max_cores, 'total number of cores assigned by the user ({}) are larger than the maxcores ({})'.format(self.used_cores, self.max_cores)
                


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#NEORL team thanks stable-baselines as we have used their own implementation of different RL 
#algorathims to establish NEORL optimizers. We have used the files in this open-source repo:

#https://github.com/hill-a/stable-baselines


import numpy as np
import tensorflow as tf
from gym.spaces import Discrete, Box, MultiBinary, MultiDiscrete


def observation_input(ob_space, batch_size=None, name='Ob', scale=False):
    """
    Build observation input with encoding depending on the observation space type

    When using Box ob_space, the input will be normalized between [1, 0] on the bounds ob_space.low and ob_space.high.

    :param ob_space: (Gym Space) The observation space
    :param batch_size: (int) batch size for input
                       (default is None, so that resulting input placeholder can take tensors with any batch size)
    :param name: (str) tensorflow variable name for input placeholder
    :param scale: (bool) whether or not to scale the input
    :return: (TensorFlow Tensor, TensorFlow Tensor) input_placeholder, processed_input_tensor
    """
    if isinstance(ob_space, Discrete):
        observation_ph = tf.placeholder(shape=(batch_size,), dtype=tf.int32, name=name)
        processed_observations = tf.cast(tf.one_hot(observation_ph, ob_space.n), tf.float32)
        return observation_ph, processed_observations

    elif isinstance(ob_space, Box):
        observation_ph = tf.placeholder(shape=(batch_size,) + ob_space.shape, dtype=ob_space.dtype, name=name)
        processed_observations = tf.cast(observation_ph, tf.float32)
        # rescale to [1, 0] if the bounds are defined
        if (scale and
           not np.any(np.isinf(ob_space.low)) and not np.any(np.isinf(ob_space.high)) and
           np.any((ob_space.high - ob_space.low) != 0)):

            # equivalent to processed_observations / 255.0 when bounds are set to [255, 0]
            processed_observations = ((processed_observations - ob_space.low) / (ob_space.high - ob_space.low))
        return observation_ph, processed_observations

    elif isinstance(ob_space, MultiBinary):
        observation_ph = tf.placeholder(shape=(batch_size, ob_space.n), dtype=tf.int32, name=name)
        processed_observations = tf.cast(observation_ph, tf.float32)
        return observation_ph, processed_observations

    elif isinstance(ob_space, MultiDiscrete):
        observation_ph = tf.placeholder(shape=(batch_size, len(ob_space.nvec)), dtype=tf.int32, name=name)
        processed_observations = tf.concat([
            tf.cast(tf.one_hot(input_split, ob_space.nvec[i]), tf.float32) for i, input_split
            in enumerate(tf.split(observation_ph, len(ob_space.nvec), axis=-1))
        ], axis=-1)
        return observation_ph, processed_observations

    else:
        raise NotImplementedError("Error: the model does not support input space of type {}".format(
            type(ob_space).__name__))



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Sun Jun 14 13:45:54 2020
#
#@author: Majdi and Paul
#"""

import random
import math
import numpy as np
import copy
import joblib
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

class SA:
    """
    Parallel Simulated Annealing
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param cooling: (str) cooling schedule, choose ``fast``, ``boltzmann``, ``cauchy``,``equilibrium``. The ``equilibrium`` mode is only valid with ``ncores > 1`` (See **Notes** below)
    :param chain_size: (int) number of individuals to evaluate in the chain every generation (e.g. like ``npop`` for other algorithms)
    :param Tmax: (int) initial/maximum temperature
    :param Tmin: (int) final/minimum temperature
    :param chi: (float or list of floats) probability of perturbing every attribute of the input ``x``, ONLY used if ``move_func=None``. 
                For ``ncores > 1``, if a scalar is provided, constant value is used across all ``ncores``. If a list of size ``ncores``
                is provided, each core/chain uses different value of ``chi`` (See **Notes** below)
    :param move_func: (function) custom self-defined function that controls how to perturb the input space during annealing (See **Notes** below)
    :param reinforce_best: (str) an option to control the starting individual of the chain at every generation. Choose ``None``, ``hard``, ``soft`` (See **Notes** below).
    :param lmbda: (float) ONLY used if ``cooling = equilibrium``, control the cooling rate, and the speed at which the algorithm converges.
    :param alpha: (float) ONLY used if ``cooling = equilibrium``, control the initial temperature of the cooling schedule.
    :param threshold: (float) ONLY used if ``cooling = equilibrium``. The threshold (in %) for the acceptance rate of solution under which the algorithm stops running. 
    :param ncores: (int) number of parallel processors (``ncores > 1`` is required for ``cooling = equilibrium``)
    :param seed: (int) random seed for sampling
    """
    def __init__ (self, mode, bounds, fit, cooling='fast', chain_size=10,  
                  Tmax=10000, Tmin=1, chi=0.1, move_func=None, reinforce_best='soft', 
                  lmbda = 1.5, alpha = 1.5, threshold = 10 ,ncores=1, seed=None):  

        set_neorl_seed(seed)
        
        self.seed=seed
        #--mir
        self.mode=mode
        if mode == 'max':
            self.fit=fit
        elif mode == 'min':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
            
        self.bounds=bounds
        self.reinforce_best=reinforce_best
        self.ncores=ncores
        self.npop=chain_size
        self.threshold = threshold
        #assert npop % self.ncores == 0, 'The number of population (npop) to run must be divisible by ncores, {} mod {} != 0'.format(npop,self.ncores)
        self.Tmax=Tmax
        self.Tmin=Tmin
        if isinstance(chi, list):
            assert len(chi) == ncores, 'The list of chi values ({}) MUST equal ncores ({})'.format(len(chi),self.ncores) 
            self.chi=chi
        elif type(chi) == float or type(chi) == int:
            self.chi=[chi]*ncores
        else:
            raise Exception ('for chi, either list of floats or scalar float are allowed')
        
        
        assert cooling in ['fast', 'boltzmann', 'cauchy', 'equilibrium'], '--error: invalid cooling is provided'
        assert reinforce_best in ['soft', 'hard', None], '--error: invalid `reinforce_best` option is provided'
        
        self.cooling=cooling 
        self.equilib_deactivate=False
        if self.cooling == 'equilibrium' and self.ncores == 1:#Paul. The equilibrium cooling is only available with multiple chains working at the same time.
            print("-- warning: equilibrium cooling is implemented ONLY for ncores > 1. The cooling is changed to default cooling --> 'fast'")
            self.cooling = 'fast'
            self.equilib_deactivate=True
        if not self.cooling=='equilibrium':# Paul
            self.T=Tmax #initialize T
        else:
            self.lmbda = lmbda
            self.alpha = alpha
        self.move_func=move_func
        if self.move_func is None:
            self.move=self.def_move
        else:
            self.move=move_func
            
            
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=[self.bounds[item][1] for item in self.bounds]
        self.ub=[self.bounds[item][2] for item in self.bounds]
            
    def GenInd(self, bounds):
        #"""
        #Individual generator
        #Input: 
        #    -bounds (dict): input paramter type and lower/upper bounds in dictionary form
        #Returns: 
        #    -individual (list): individual position
        #"""
        content=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                content.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                content.append(random.uniform(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'grid':
                content.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        ind=list(content)
        return ind

    def sampler(self,bound):
        #"""
        #This function takes input as [type, low, high] and returns a sample  
        #This is to sample the input parameters during optimisation
        #"""
        if bound[0] == 'int':
            sample=random.randint(bound[1], bound[2])
        elif bound[0] == 'float':
           sample=random.uniform(bound[1], bound[2])
        elif bound[0] == 'grid':
            sample=random.sample(bound[1],1)[0]
        else:
            raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')
                
        return sample

    def ensure_bounds(self, vec): # bounds check

        vec_new = []

        for i, (key, val) in enumerate(self.bounds.items()):
            # less than minimum 
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
            # more than maximum
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
            # fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
        
        return vec_new

    def fit_worker(self, x):
        #This worker is for parallel calculations of the GWO
        
        # Clip the wolf with position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x)
        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        
        return fitness
    
    def def_move(self, x, chi):
        #"""
        #This function is to perturb x attributes by probability chi
        #Inputs:
        #    x: input vector 
        #    chi: perturbation probablity between 0 and 1
        #Returns: perturbed vector x
        #"""
        i=0
        x_new=x.copy()
        for item in self.bounds:
            if random.random() < chi:
                sample = self.sampler(self.bounds[item])
                while x_new[i] == sample and (self.bounds[item][1] != self.bounds[item][2]): 
                    sample = self.sampler(self.bounds[item])
                x_new[i] = sample
            i+=1
        return x_new
        
    def temp(self,step):
        #"""
        # Function to anneal temperature
        #"""

        if self.cooling=='fast':
            Tfac = -np.log(float(self.Tmax) / self.Tmin)
            T = self.Tmax * np.exp( Tfac * step / self.steps)
        elif self.cooling=='boltzmann':
            T = self.Tmax / np.log(step + 1)
        elif self.cooling=='cauchy':
            T = self.Tmax / (step + 1)
        # Paul
        elif self.cooling=='equilibrium':
            accept_rate = np.mean(self.accepts)
            sigma = np.std(self.accepted_energy)
            g_rho = 4 * accept_rate * (1 - accept_rate)**2 / ((2 - accept_rate)**2)
            T = np.max([1 / (1 / self.T + self.lmbda * (1 / sigma) * (self.T / sigma)**2 * g_rho),self.Tmin])
        else:
            raise Exception ('--error: unknown cooling mode is entered, fast, boltzmann, cauchy, or equilibrium are ONLY allowed')
        
        return T
    
    def chain_object(self,inp):
        #"""
        #This function is a multiprocessing object, used to be passed to Pool, that respresents 
        #an individual SA chain. 
        #Input:
        #    inp: a list contains the following inputs in order
        #    inp[0] --> x0: initial guess to chain 
        #    inp[1] --> E0: initial energy of x0
        #    inp[2] --> min_step: min step to start this chain 
        #    inp[3] --> max_step: max step to terminate this chain 
        #    inp[4] --> core_seed: seed for this chain
        #returns: 
        #    x_best, E_best: best obtained from this chain
        #    T: last temperature for this chain
        #    accepts, rejects, improves for this chain
        #"""
        x_prev=copy.deepcopy(inp[0])
        x_best=copy.deepcopy(x_prev)
        E_prev=inp[1]
        E_best=inp[1]
        min_step=inp[2]
        max_step=inp[3]
        core_seed=inp[4]
        
        if not (self.seed is None):
            random.seed(self.seed + core_seed)
            np.random.seed(self.seed + core_seed)
        
        rejects=0; accepts=0; improves=0
        k=min_step
        
        if self.cooling == 'equilibrium':#Paul.
            T = copy.deepcopy(self.T)#Paul
            accepted_energy = []#Paul
        while k <= max_step:
            if self.cooling != 'equilibrium': #Paul. if updated here may not return nan in first few steps
                T=self.temp(step=k)
                
            if self.move_func is None:
                x=copy.deepcopy(self.move(x=x_prev,chi=self.chi[core_seed-1]))
            else:
                x=copy.deepcopy(self.move(x=x_prev))
            
            x=self.ensure_bounds(x)
            #SA is programmed to maximize reward
            E=self.fit_worker(x)
            dE = E - E_prev        
            #-----------------------------------
            # Improve/Accept/Reject
            #-----------------------------------
            if dE > 0: #improvement
                improves += 1
                accepts += 1
                x_prev = copy.deepcopy(x)
                E_prev = E
                if E > E_best:
                    x_best = copy.deepcopy(x)
                    E_best = E     
                if self.cooling == 'equilibrium':# Paul
                    accepted_energy.append(E)
            elif np.exp(dE/T) >= random.random(): #accept the state
                accepts += 1
                x_prev = copy.deepcopy(x)
                E_prev = E
                if self.cooling == 'equilibrium':# Paul
                    accepted_energy.append(E)
            else:
                # Reject the new solution (maintain the current state!)
                rejects += 1
            k+=1
        if self.cooling == 'equilibrium':#Paul.
            return x_prev, E_prev, T, accepts, rejects, improves, x_best, E_best,accepted_energy
        else:
            return x_prev, E_prev, T, accepts, rejects, improves, x_best, E_best
        
    def chain(self, x0, E0, step0):
        #"""
        #This function creates ``ncores`` independent SA chains with same initial guess x0, E0 and 
        #runs them via multiprocessing Pool.
        #Input:
        #    x0: initial input guess (comes from previous annealing chains or from replay memory)
        #    E0: energy/fitness value of x0
        #    step0: is the first time step to use for temperature annealing
        #    npop: total number of individuals to be evaluated in this annealing stage
        #returns: 
        #    x_best, E_best, and T obtained from this annealing stage from all chains
        #"""
        
        #Append and prepare the input list to be passed as input to Pool.map
        core_list=[]
        core_step_min=step0
        for j in range(1,self.ncores+1):
            core_step_max=step0+j*self.npop-1
            core_list.append([x0[j-1], E0[j-1], core_step_min, core_step_max, j])
            core_step_min=core_step_max+1
            
        
        if self.ncores > 1:

            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                results=parallel(joblib.delayed(self.chain_object)(item) for item in core_list)
        else:
            results=[]
            results.append(list(self.chain_object(core_list[0])))
                
        # Determine the index and the best solution from all chains
        #best_index=[y[0] for y in results].index(min([y[0] for y in results]))
        self.x_last=[item[0] for item in results]
        self.E_last=[item[1] for item in results]
        self.T=results[-1][2] # get the temperature of the last chain
        self.accepts=[np.round(item[3]/self.npop*100,1) for item in results] #convert to rate
        self.rejects=[np.round(item[4]/self.npop*100,1) for item in results] #convert to rate
        self.improves=[np.round(item[5]/self.npop*100,1) for item in results] #convert to rate
        
        if self.cooling == 'equilibrium':#Paul
            self.accepted_energy = [item[8] for item in results]#Paul
            self.accepted_energy = [elem for energy in self.accepted_energy for elem in energy]
            if np.std(self.accepted_energy) == 0: # prevent the std to be 0
                pass
            else:
                self.T=self.temp(step=None)
            
        self.x_best, self.E_best=[item[6] for item in results], [item[7] for item in results]
        
        return self.x_last, self.E_last, self.T, self.accepts, self.rejects, self.improves, self.x_best, self.E_best
    
    def InitChains(self, x0=None):
        
        #initialize the chain and run them in parallel (these samples will be used to initialize the annealing process)
        #Establish the chain
        if x0:
            if self.ncores==1:  
                if not any(isinstance(el, list) for el in x0):  #ensure a list of list is submitted for ncores=1
                    x0=[x0]
            assert len(x0) == self.ncores, '--error: the length of x0 ({}) MUST equal the number of ncores/parallel chains ({})'.format(len(x0), self.ncores)

            for i in range(len(x0)):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    x0[i] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)       
        
        else:
            x0=[]
            for i in range (self.ncores):
                x0.append(self.GenInd(self.bounds))
        
        #Evaluate the swarm
        if self.ncores > 1:  #evaluate chain in parallel
            core_list=[]
            for ind in x0:
                core_list.append(ind)
           
            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                E0=parallel(joblib.delayed(self.fit_worker)(item) for item in core_list)
            if self.cooling == 'equilibrium': # Paul
                self.T = np.max([self.alpha * np.std(E0),self.Tmin,1e-10])

        else: #evaluate swarm in series
            E0=[]
            for ind in x0:
                fitness=self.fit_worker(ind)
                E0.append(fitness)
        

        return x0, E0 #return initial guess and initial fitness      
    
    def evolute(self, ngen, x0=None, verbose=False):
        """
        This function evolutes the SA algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial samples to start the evolution (``len(x0)`` must be equal to ``ncores``)
        :param verbose: (int) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        #chain statistics
        #self.accept=0
        #self.reject=0
        #self.improve=0
        
        stat={'x':[], 'fitness':[], 'T':[], 'accept':[], 'reject':[], 'improve':[]}
        E_opt=-np.inf
        
        step0=1
        self.steps=ngen * self.npop
        
        if x0:
            if isinstance(x0[0], list):
                if not all(len(item) == len(x0[0]) for item in x0):
                    raise Exception ('--error: the variable x0 must be a list of lists, and all internal lists must have same length.')
            else:
                x0=[x0]
                
            assert len(x0) == self.ncores, '--error: Length of initial guesses x0 ({}) for chains do not equal to ncores or # of chains ({})'.format(len(x0), self.ncores)
            assert len(x0[0]) == len(self.bounds), '--error: Length of every list in x0 ({}) do not equal to the size of parameter space in bounds ({})'.format(len(x0[0]), len(self.bounds))
            xinit, Einit=self.InitChains(x0=x0)
        else:
            xinit, Einit=self.InitChains()
            
        x_next=copy.deepcopy(xinit)
        E_next=copy.deepcopy(Einit)
        
        ngen=int(ngen/self.ncores)
        for i in range (ngen):
            #if self.cooling == 'equilibrium':# Paul
            #    self.accepted_energy = [] # initialize list of accepted energy to empty at step i
            x_next,E_next,self.T, acc, rej, imp, x_best, E_best=self.chain(x0=x_next, E0=E_next, step0=step0)
            step0=step0+self.npop*self.ncores
            arg_max=np.argmax(E_best)
            stat['x'].append(x_best[arg_max])
            if self.mode=='max':
                stat['fitness'].append(max(E_best))
            else:
                stat['fitness'].append(-max(E_best))
            stat['T'].append(self.T)
            stat['accept'].append(acc[arg_max])
            stat['reject'].append(rej[arg_max])
            stat['improve'].append(imp[arg_max])
            
            if max(E_best) > E_opt:
                E_opt=max(E_best)
                x_opt=copy.deepcopy(x_best[arg_max])
            
            if self.cooling == 'equilibrium': # Paul
                if np.mean(self.accepts) <= self.threshold: # help prevent the std to be 0
                    print("--warning: The SA stopped because the average acceptance rate throughout the chain {} % falls below the threshold {} %".format(np.mean(self.accepts),self.threshold))
                    break
                elif np.mean(self.accepts) == 0:
                    print("--warning: The SA stopped because the average acceptance rate throughout the chain {} % reaches 0%".format(np.mean(self.accepts),self.threshold))
                    break
            # Paul
            if self.reinforce_best == 'hard':
                x_next=[x_opt]*self.ncores
                E_next=[E_opt]*self.ncores
            elif self.reinforce_best == 'soft':
                sampling = np.zeros(self.ncores)
                normalization = 1 / np.sum(np.exp(- np.abs(E_next) / self.T)) # cte to generate a probability
                sampling[0] = np.exp(- abs(E_next[0]) / self.T) # utilize last accepted solution to generate the sampling
                if self.ncores != 1:
                    rho = np.random.uniform(0,1,size = self.ncores) # sample random number between 0 and 1
                else:
                    rho = [1] # probability of choosing itself is 1    
                for i in range(1,self.ncores):
                    sampling[i] = sampling[i - 1] + np.exp(- abs(E_next[i]) / self.T)
                sampling = sampling * normalization
                for count,prob in enumerate(rho):
                    if prob == 1:
                        index = self.ncores - 1
                    elif math.isnan(sampling[count]):# exponantial can lead to numerical erros
                        index = np.argmax(E_next)
                    else:
                        if prob <= sampling[0]:
                            index = 0
                        else:
                            index = np.where(sampling > rho)[0][0]
                    x_next[count] = x_best[index] # re-initialize with the best ever found by the index'th Markov Chain
                    E_next[count] = E_best[index]

            #mir-grid
            if self.grid_flag:
                x_opt_correct = decode_discrete_to_grid(x_opt, self.orig_bounds, self.bounds_map)
            else:
                x_opt_correct = x_opt
                    
            if verbose:
                print('***********************************************************************')
                print('SA step {}/{}, T={}, Ncores={}, Cooling={}, Reinforce={}'.format(step0-1,self.steps,np.round(self.T), self.ncores, self.cooling, self.reinforce_best))
                print('***********************************************************************')
                print('Statistics for the {} parallel chains'.format(self.ncores))
                if self.mode=='max': 
                    print('Best fitness:', np.round(max(E_best),6))
                else: 
                    print('Best fitness:', -np.round(max(E_best),6))
                print('Best individual:', x_opt_correct)
                print('Acceptance Rate (%):', acc)
                print('Rejection Rate (%):', rej)
                print('Improvment Rate (%):', imp)
                print('***********************************************************************')

            
        #--mir
        if self.mode=='max':
            self.E_opt_correct=E_opt
        else:
            self.E_opt_correct=-E_opt
                
        if verbose:
            print('------------------------ SA Summary --------------------------')
            print('Best fitness (y) found:', self.E_opt_correct)
            print('Best individual (x) found:', x_opt_correct)
            print('--------------------------------------------------------------')

        if self.equilib_deactivate: #Paul.
            print("-- warning: equilibrium cooling is implemented ONLY for ncores > 1. The cooling is changed to default cooling --> 'fast'")
            
        return x_opt_correct, self.E_opt_correct, stat



from neorl import ACKDE
from neorl import CreateEnvironment

def test_ackde():
    def Sphere(individual):
        """Sphere test objective function.
                F(x) = sum_{i=1}^d xi^2
                d=1,2,3,...
                Range: [-100,100]
                Minima: 0
        """
        y=sum(x**2 for x in individual)
        return y
    
    
    #Setup the parameter space (d=5)
    nx=5
    BOUNDS={}
    for i in range(1,nx+1):
        BOUNDS['x'+str(i)]=['float', -100, 100]
    
    #create an enviroment class for RL/ACKTR
    env=CreateEnvironment(method='acktr', fit=Sphere, ncores=1,  
                          bounds=BOUNDS, mode='min', episode_length=50)
    
    #change hyperparameters of ACKTR/DE if you like (defaults should be good to start with)
    h={'F': 0.5,
       'CR': 0.3,
       'n_steps': 20,
       'learning_rate': 0.001}
    
    #Important: `mode` in CreateEnvironment and `mode` in ACKDE must be consistent
    #fit is needed to be passed again for DE, must be same as the one used in env
    ackde=ACKDE(mode='min', fit=Sphere, npop=60,
                env=env, npop_rl=6, init_pop_rl=False, 
                bounds=BOUNDS, hyperparam=h, seed=1)
    #first run RL for some timesteps
    rl=ackde.learn(total_timesteps=2000, verbose=True)
    #second run DE, which will use RL data for guidance
    ackde_x, ackde_y, ackde_hist=ackde.evolute(ngen=100, ncores=1, verbose=True) #ncores for DE
    
    return

test_ackde()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from neorl.evolu.es import ES

class ESTUNE:
    """
    A module for Evolutionary search for hyperparameter tuning based on ES algorithm
    
    :param param_grid: (dict) the type and range of each hyperparameter in a dictionary form (types are ``int/discrete`` or ``float/continuous`` or ``grid/categorical``).
    :param fit: (function) the self-defined fitness function that includes the hyperparameters as input and algorithm score as output
    :param mode: (str) problem type, either ``min`` for minimization problem 
                 or ``max`` for maximization. Default: Evolutionary tuner 
                 is set to maximize an objective
    :param ngen: (int) number of ES generations to run, total number of hyperparameter tests is ``ngen * 10`` (see **Notes** for an important remark) 
    :param seed: (int) random seed for sampling reproducibility
    """
    def __init__(self, param_grid, fit, mode='max', ngen=10, seed=None):
        #--mir
        self.mode=mode

        self.fit=fit
        assert self.mode in ['min', 'max'], '--error: The mode entered by user is invalid, use either `min` or `max`'
        self.param_grid=param_grid
        self.seed=seed
        self.npop=10
        #if self.ncases < self.npop:
        #    self.ncases = self.npop 
        #    print('--warning: ncases={ncases} < {npop} is given by the user, but ncases must be more than {npop}, reset ncases to {npop}'.format(ncases=self.ncases, npop=self.npop))  
            
        self.ngen = ngen
        self.param_names=[item for item in self.param_grid]

    def plot_results(self, pngname='evolu_tune'):
        if self.mode=='max':
            plt.plot(pd.DataFrame.cummax(self.evolures['score']), '-og')
            plt.ylabel('Max score so far')
        else:
            plt.plot(pd.DataFrame.cummin(self.evolures['score']), '-og')
            plt.ylabel('Min score so far')
            
        plt.xlabel('Generation # (Every generation includes {} cases)'.format(self.npop))
        plt.grid()
        if pngname is not None:
            plt.savefig(str(pngname)+'.png', dpi=200, format='png')
        plt.close()


    def tune(self, ncores=1, csvname=None, verbose=True):
        """
        This function starts the tuning process with specified number of processors
    
        :param ncores: (int) number of parallel processors.
        :param csvname: (str) the name of the csv file name to save the tuning results (useful for expensive cases as the csv file is updated directly after the case is done)
        :param verbose: (bool) whether to print updates to the screen or not
        """
        self.ncores=ncores
        self.csvlogger=csvname
        self.verbose=verbose

        if self.verbose:
            print('***********************************************************************')
            print('****************Evolutionary Search is Running*************************')
            print('***********************************************************************')
            
            if self.ncores > 1:
                print('--- Running in parallel with {} cores'.format(self.ncores))
                
        es=ES(mode=self.mode, bounds=self.param_grid, fit=self.fit, 
              lambda_=self.npop, mu=self.npop, mutpb=0.25,
             cxmode='cx2point', cxpb=0.7, ncores=self.ncores, seed=self.seed)
        x_best, y_best, es_hist=es.evolute(ngen=self.ngen, verbose=self.verbose)

        self.evolures = pd.DataFrame(es.best_indvs, columns=self.param_names)
        self.evolures.index += 1
        self.evolures['score'] = es_hist['local_fitness']

        if self.csvlogger:
            self.evolures.index.name='id'
            self.evolures.to_csv(self.csvlogger)
                
        return self.evolures


from neorl import DE

def test_de():
    #Define the fitness function
    def FIT(individual):
            """Sphere test objective function.
                    F(x) = sum_{i=1}^d xi^2
                    d=1,2,3,...
                    Range: [-100,100]
                    Minima: 0
            """
            y=sum(x**2 for x in individual)
            return y
    
    #Setup the parameter space (d=5)
    nx=5
    BOUNDS={}
    for i in range(1,nx+1):
            BOUNDS['x'+str(i)]=['float', -100, 100]
    
    de=DE(mode='min', bounds=BOUNDS, fit=FIT, npop=60, F=0.5, CR=0.7, ncores=1, seed=1)
    x_best, y_best, de_hist=de.evolute(ngen=100, verbose=0)

test_de()


from abc import ABC, abstractmethod

import numpy as np


class AdaptiveParamNoiseSpec(object):
    """
    Implements adaptive parameter noise

    :param initial_stddev: (float) the initial value for the standard deviation of the noise
    :param desired_action_stddev: (float) the desired value for the standard deviation of the noise
    :param adoption_coefficient: (float) the update coefficient for the standard deviation of the noise
    """

    def __init__(self, initial_stddev=0.1, desired_action_stddev=0.1, adoption_coefficient=1.01):
        self.initial_stddev = initial_stddev
        self.desired_action_stddev = desired_action_stddev
        self.adoption_coefficient = adoption_coefficient

        self.current_stddev = initial_stddev

    def adapt(self, distance):
        """
        update the standard deviation for the parameter noise

        :param distance: (float) the noise distance applied to the parameters
        """
        if distance > self.desired_action_stddev:
            # Decrease stddev.
            self.current_stddev /= self.adoption_coefficient
        else:
            # Increase stddev.
            self.current_stddev *= self.adoption_coefficient

    def get_stats(self):
        """
        return the standard deviation for the parameter noise

        :return: (dict) the stats of the noise
        """
        return {'param_noise_stddev': self.current_stddev}

    def __repr__(self):
        fmt = 'AdaptiveParamNoiseSpec(initial_stddev={}, desired_action_stddev={}, adoption_coefficient={})'
        return fmt.format(self.initial_stddev, self.desired_action_stddev, self.adoption_coefficient)


class ActionNoise(ABC):
    """
    The action noise base class
    """

    def __init__(self):
        super(ActionNoise, self).__init__()

    def reset(self) -> None:
        """
        call end of episode reset for the noise
        """
        pass

    @abstractmethod
    def __call__(self) -> np.ndarray:
        raise NotImplementedError()


class NormalActionNoise(ActionNoise):
    """
    A Gaussian action noise

    :param mean: (float) the mean value of the noise
    :param sigma: (float) the scale of the noise (std here)
    """

    def __init__(self, mean, sigma):
        super().__init__()
        self._mu = mean
        self._sigma = sigma

    def __call__(self) -> np.ndarray:
        return np.random.normal(self._mu, self._sigma)

    def __repr__(self) -> str:
        return 'NormalActionNoise(mu={}, sigma={})'.format(self._mu, self._sigma)


class OrnsteinUhlenbeckActionNoise(ActionNoise):
    """
    A Ornstein Uhlenbeck action noise, this is designed to approximate brownian motion with friction.

    Based on http://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab

    :param mean: (float) the mean of the noise
    :param sigma: (float) the scale of the noise
    :param theta: (float) the rate of mean reversion
    :param dt: (float) the timestep for the noise
    :param initial_noise: ([float]) the initial value for the noise output, (if None: 0)
    """

    def __init__(self, mean, sigma, theta=.15, dt=1e-2, initial_noise=None):
        super().__init__()
        self._theta = theta
        self._mu = mean
        self._sigma = sigma
        self._dt = dt
        self.initial_noise = initial_noise
        self.noise_prev = None
        self.reset()

    def __call__(self) -> np.ndarray:
        noise = self.noise_prev + self._theta * (self._mu - self.noise_prev) * self._dt + \
                self._sigma * np.sqrt(self._dt) * np.random.normal(size=self._mu.shape)
        self.noise_prev = noise
        return noise

    def reset(self) -> None:
        """
        reset the Ornstein Uhlenbeck noise, to the initial position
        """
        self.noise_prev = self.initial_noise if self.initial_noise is not None else np.zeros_like(self._mu)

    def __repr__(self) -> str:
        return 'OrnsteinUhlenbeckActionNoise(mu={}, sigma={})'.format(self._mu, self._sigma)



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
"""
Created on Sun Jun 14 13:45:54 2020

@author: Majdi
"""

import random
import numpy as np
import copy
from neorl.hybrid.pesacore.er import ExperienceReplay
import time
import multiprocessing
import multiprocessing.pool
import joblib
from neorl.utils.seeding import set_neorl_seed
class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

class SAMod(ExperienceReplay):
    
    def __init__ (self, bounds, fit, steps, memory=None, ncores=1, chi=0.1, replay_rate=0, cooling='fast', Tmax=10000, Tmin=1, seed=None):  
        """
        Parallel SA:
        A Synchronous Approach with Occasional Enforcement of Best Solution-Fixed Intervals
        Inputs:
            bounds (dict): input paramter lower/upper bounds in dictionary form
            memory (class): a class object created by ExperienceReplay
            fit (function): fitness function 
            ncores (int): parallel cores
            chi (float/list): a float or list of floats representing perturbation probablity for each parallel chain 
                 if float, one chi used for all chains 
                 else a list of chi's with size ``ncores`` are used for each chain
            cooling (str): cooling schedule, either fast, boltzmann, or cauchy are allowed
            Tmax (int): maximum temperature
            Tmin (int): minimum temperature
        """
        
        set_neorl_seed(seed)
        
        if memory:
            self._memory=memory
        else:
            self._memory=False
            
        self.batch_size=10
        self.replay_rate=replay_rate
        self.bounds=bounds
        self.ncores=ncores
        self.Tmax=Tmax
        self.Tmin=Tmin
        if type(chi) == list:
            assert len(chi) == ncores, 'The list of chi values ({}) MUST equal ncores ({})'.format(len(chi),self.ncores) 
            self.chi=chi
        elif type(chi) == float or type(chi) == int:
            self.chi=[chi]*ncores
        else:
            raise Exception ('for chi, either list of floats or scalar float are allowed')
            
        self.cooling=cooling
        self.fit=fit
        self.steps=steps
        self.T=Tmax #initialize T
        
        #chain statistics
        self.accept=0
        self.reject=0
        self.improve=0

    def sampler(self,bound):
        """
        This function takes input as [type, low, high] and returns a sample  
        This is to sample the input parameters during optimisation
        """
        if bound[0] == 'int':
            sample=random.randint(bound[1], bound[2])
        elif bound[0] == 'float':
           sample=random.uniform(bound[1], bound[2])
        elif bound[0] == 'grid':
            sample=random.sample(bound[1],1)[0]
        else:
            raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')
                
        return sample
    
    def move(self, x, chi):
        """
        This function is to perturb x attributes by probability chi
        Inputs:
            x: input vector 
            chi: perturbation probablity between 0 and 1
        Returns: perturbed vector x
        """
        i=0
        x_new=x.copy()
        for item in self.bounds:
            if random.random() < chi:
                sample = self.sampler(self.bounds[item])
                while x_new[i] == sample and (self.bounds[item][1] != self.bounds[item][2]): 
                    sample = self.sampler(self.bounds[item])
                x_new[i] = sample
            i+=1
        return x_new
        
    def temp(self,step):
        
        
        if self.cooling=='fast':
            Tfac = -np.log(float(self.Tmax) / self.Tmin)
            T = self.Tmax * np.exp( Tfac * step / self.steps)
        elif self.cooling=='boltzmann':
            T = self.Tmax / np.log(step + 1)
        elif self.cooling=='cauchy':
            T = self.Tmax / (step + 1)
        else:
            raise Exception ('--error: unknown cooling mode is entered, fast, boltzmann, or cauchy are ONLY allowed')
        
        return T
    
    def chain_object (self,inp):
        """
        This function is a multiprocessing object, used to be passed to Pool, that respresents 
        an individual SA chain. 
        Input:
            inp: a list contains the following inputs in order
            inp[0] --> x0: initial guess to chain 
            inp[1] --> E0: initial energy of x0
            inp[2] --> min_step: min step to start this chain 
            inp[3] --> max_step: max step to terminate this chain 
            inp[4] --> core_seed: seed for this chain
        returns: 
            x_best, E_best: best obtained from this chain
            T: last temperature for this chain
            accepts, rejects, improves for this chain
        """
        
        x_prev=copy.deepcopy(inp[0])
        x_best=copy.deepcopy(x_prev)
        E_prev=inp[1]
        E_best=inp[1]
        min_step=inp[2]
        max_step=inp[3]
        core_seed=inp[4]
        random.seed(core_seed)
        
        rejects=0; accepts=0; improves=0
        k=min_step
        while k <= max_step:
            T=self.temp(step=k)
            
            if random.random() < self.replay_rate and self._memory: #replay memory
                x, E, _=self._memory.sample(batch_size=1, mode='greedy', seed=core_seed)[0]
                E=self.fit(x)
                #if core_seed==1 or core_seed==10:
                #    print('memory sample {}, step {}, E {}'.format(core_seed, k, np.round(E,2)))
            else: #random-walk
                x=copy.deepcopy(self.move(x=x_prev,chi=self.chi[core_seed-1]))
                #SA is programmed to maximize reward
                E=self.fit(x)
            
            dE = E - E_prev        
            #-----------------------------------
            # Improve/Accept/Reject
            #-----------------------------------
            if dE > 0: #improvement
                improves += 1
                accepts += 1
                x_prev = copy.deepcopy(x)
                E_prev = E
                if E > E_best:
                    x_best = copy.deepcopy(x)
                    E_best = E     
                
            elif np.exp(dE/T) >= random.random(): #accept the state
                accepts += 1
                x_prev = copy.deepcopy(x)
                E_prev = E
            else:
                # Reject the new solution (maintain the current state!)
                rejects += 1
            k+=1
        
            #debugging lines
            #if core_seed==1 or core_seed==10:
            #    print('E_prev '+str(core_seed), np.round(E_prev,2))
            #    print('memory_SA', len(self._memory.storage))
            #assert self.fit(x_prev) == E_prev
        return x_prev, E_prev, T, accepts, rejects, improves, x_best, E_best
        
    def chain(self, x0, E0, step0, npop):
        """
        This function creates ``ncores`` independent SA chains with same initial guess x0, E0 and 
        runs them via multiprocessing Pool.
        Input:
            x0: initial input guess (comes from previous annealing chains or from replay memory)
            E0: energy/fitness value of x0
            step0: is the first time step to use for temperature annealing
            npop: total number of individuals to be evaluated in this annealing stage
        returns: 
            x_best, E_best, and T obtained from this annealing stage from all chains
        """
        assert npop % self.ncores == 0, 'The number of communications to run must be divisible by ncores, {} mod {} != 0'.format(npop,self.ncores)
        core_npop =int(npop/self.ncores)
        
        #Append and prepare the input list to be passed as input to Pool.map
        core_list=[]
        core_step_min=step0
        for j in range(1,self.ncores+1):
            core_step_max=step0+j*core_npop-1
            core_list.append([x0[j-1], E0[j-1], core_step_min, core_step_max, j])
            core_step_min=core_step_max+1
        
        if self.ncores > 1:
            # create and run the Pool
            t0=time.time()
            p=MyPool(self.ncores)
            results = p.map(self.chain_object, core_list)
            p.close()
            p.join()
            
            #with joblib.Parallel(n_jobs=self.ncores) as parallel:
            #    results=parallel(joblib.delayed(self.chain_object)(item) for item in core_list)
            self.partime=time.time()-t0
            #print('SA:', self.partime)
        else:
            results=[]
            results.append(list(self.chain_object(core_list[0])))
            self.partime=0
                
        # Determine the index and the best solution from all chains
        #best_index=[y[0] for y in results].index(min([y[0] for y in results]))
        self.x_last=[item[0] for item in results]
        self.E_last=[item[1] for item in results]
        self.T=results[-1][2] # get the temperature of the last chain
        self.accepts=[np.round(item[3]/core_npop*100,1) for item in results] #convert to rate
        self.rejects=[np.round(item[4]/core_npop*100,1) for item in results] #convert to rate
        self.improves=[np.round(item[5]/core_npop*100,1) for item in results] #convert to rate
        
        self.x_best, self.E_best=[item[6] for item in results], [item[7] for item in results]
        
        return self.x_last, self.E_last, self.T, self.accepts, self.rejects, self.improves, self.x_best, self.E_best
    
    def anneal(self, ngen, npop, x0, E0, step0, verbose=0):
        """
        Perform annealing over total ``steps`` by updating chains every ``npop``
        Returns the best ``x`` and ``energy`` over the whole stage 
        """
        assert len(x0) == self.ncores, 'Length of initial guesses x0 ({}) for chains do not equal to ncores or # of chains ({})'.format(len(x0), self.ncores)
        assert len(E0) == self.ncores, 'Length of initial fitness E0 ({}) for chains do not equal to ncores or # of chains ({})'.format(len(E0), self.ncores)
        x_next=copy.deepcopy(x0)
        E_next=copy.deepcopy(E0)
        steps=ngen*npop
        for i in range (ngen):
            x_next,E_next,self.T, acc, rej, imp, x_best, E_best=self.chain(x0=x_next, E0=E_next, step0=step0, npop=npop)
            step0=step0+npop
            arg_max=np.argmax(E_best)
            if verbose:
                print('************************************************************')
                print('SA step {}/{}, T={}'.format(step0-1,self.steps,np.round(self.T)))
                print('************************************************************')
                print('Statistics for the {} parallel chains'.format(self.ncores))
                print('Best fitness:', np.round(E_best,2))
                print('Best individual:', x_best[arg_max])
                print('Acceptance Rate (%):', acc)
                print('Rejection Rate (%):', rej)
                print('Improvment Rate (%):', imp)
                print('************************************************************')
            
        return x_next, E_next, self.T, acc, rej, imp, x_best, E_best, self.partime


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

"""
Created on Tue Feb 25 12:09:27 2020

@author: Majdi Radaideh
"""

# List of all neorl variables 


class InputParam():
    def __init__ (self):
    
        """
        #---------------------------------    
        #Key: [Default value (if any), Category, Type]
        #---------------------------------   
            
        #Default Value: is the default value of optional parameter if not specified.
        #Key: the type of the input: required ('r'), optional ('o'), required in special cases ('rs')
        #Type: data structure type: int, float, str, vec, etc. 
        
        # input keys: 
        r: required parameter, if not defined, error is raised, neorl terminates.
        o: optional parameter, if not defined, default value is used, warning is raised, execution continues.
        rs: required for special cases, special case is explained next to each parameter, if not defined and special case is violated, error is raised, neorl terminates. 
        
        """
        
        #---------------------------------
        # General Card
        #---------------------------------
        self.gen_dict={
        'neorl_mode': ['run','o','str'],  # a 'developer' option for input parser testing, run or test
        'env': [None,'r','str'],          # pre-registered env name     
        'env_data': [None,'o','dict'],     # optional env data 
        'exepath': [None,'o','str'],      # see manual
        'maxcores': [16,'o', 'int'],      # or use -1 to infer maxcores
        'daemon': [True,'o', 'bool'],      # daemon option to allow running nested children in parallel for RL, set to False if yes
        'nactions': [None,'r', 'int'],    # action size to take by agent (self.action_space) 
        'xsize':[None,'r','int'],         # see manual
        'xsize_plot':[None,'o','int'],    # see manual 
        'ysize':[None,'r','int'],         # see manual
        'ynames':[['y'], 'o', 'strvec'],  # default names x1, x2, ..., x{xsize_plot}
        'xnames':[['x'], 'o', 'strvec'],  # default names y1, y2, ..., x{ysize}
        'plot_mode':['subplot','o','str'] #either subplot (recommended) or classic 
        }
        
        #---------------------------------
        # TUNE card 
        #---------------------------------
        self.tune_dict={
        'flag': [False, 'o', 'bool'],       #this will be activated if READ TUNE is found in input 
        'ncores': [1,'o','int'],            #number of parallel tuned cases to run    
        'mode': ['run','o','str'],  # tune mode: run or test, test will generate cases but not run them. 
        'method': [None,'r','str'],  # tune method: grid search, random search, genentic algorathim
        'n_last_episodes': [50,'o','int'],  # number of last episodes to average the reward and determine convergence
        'ncases': [100,'o','int'],  # number of last episodes to average the reward and determine convergence 
        'extfiles':[None, 'o', 'strvec']
        }
        
        #---------------------------------
        # DQN Card
        #---------------------------------
        self.dqn_dict={
        'mode': ['train', 'r', 'str'],     # either train, continue, or test     
        'time_steps': [50000,'rs','int'],  # required for mode=train/continue
        'flag': [False, 'o', 'bool'],      # this will be activated if READ DQN is found in input
        'casename': ['dqn', 'o', 'str'],   # prefix for logging and results
        'ncores': [1,'o','int'],                        # only 1 core supported
        'gamma': [0.99, 'o', 'float'],                  # see manual 
        'learning_rate': [1e-3, 'o', 'float'],          # see manual
        'buffer_size': [50000, 'o', 'int'],             # see manual
        'exploration_fraction': [0.1, 'o', 'float'],    # see manual
        'exploration_final_eps': [0.02, 'o', 'float'],  # see manual
        'exploration_initial_eps': [1.0, 'o', 'float'], # see manual
        'train_freq': [1, 'o', 'float'],                # see manual
        'batch_size': [32, 'o', 'int'],                 # see manual
        'double_q': [True, 'o', 'bool'],                # see manual
        'learning_starts': [1000, 'o', 'int'],          # see manual
        'target_network_update_freq': [1000, 'o', 'int'], # see manual
        
        #Priortized Replay Parameters 
        'prioritized_replay': [False, 'o', 'bool'], # whether to use priortized replay or not
        'prioritized_replay_alpha': [0.6, 'o', 'float'], # how much prioritization is used (0 is the uniform case)
        'prioritized_replay_beta0': [0.4, 'o', 'float'], # initial value of beta for prioritized replay buffer
        
        # for testing or continue modes 
        'model_load_path': [None,'rs', 'str'], # pre-trained model required only for mode=continue/test 
        
         # for plotting 
        'check_freq': [1000,'o', 'int'],     # see manual
        'avg_episodes': [20,'o', 'int'],     # see manual
        'tensorboard': [False,'o','bool'],   # tensorboard view option, see manual for usage 
        
        # only for testing
        'n_eval_episodes': [None,'rs', 'int'],   # required only for mode=test 
        'render': [False,'o', 'bool'],           # option to plot during test via self.render() 
        }
        
        #---------------------------------
        # PPO Card
        #---------------------------------
        self.ppo_dict={
        'mode': ['train', 'r', 'str'],     # either train, continue, or test
        'time_steps': [50000,'rs','int'],  # required only for mode=train/continue 
        'flag': [False, 'o', 'bool'],      # this will be activated if READ PPO is found in input
        'casename': ['ppo', 'o', 'str'],          # prefix for logging and results 
        'ncores': [1,'o','int'],                  # multiple cores are available
        'n_steps': [128,'o','int'],               # see manual
        'gamma': [0.99, 'o', 'float'],            # see manual
        'learning_rate': [0.00025, 'o', 'float'], # see manual 
        'ent_coef': [0.01, 'o', 'float'],         # see manual
        'vf_coef': [0.5, 'o', 'float'],           # see manual
        'max_grad_norm': [0.5, 'o', 'float'],     # see manual
        'lam': [0.95, 'o', 'float'],              # see manual
        'nminibatches': [4, 'o', 'int'],          # see manual
        'noptepochs': [4, 'o', 'int'],            # see manual
        'cliprange': [0.2, 'o', 'float'],         # see manual
        'cliprange_vf': [0.2, 'o', 'float'],      # see manual
        
        # for plotting 
        'check_freq': [1000,'o', 'int'],         # see manual
        'avg_episodes': [20,'o', 'int'],         # see manual
        'tensorboard': [False,'o','bool'],       # tensorboard view option, see manual for usage 
        
        # for testing or continue modes
        'model_load_path': [None,'rs', 'str'],   # required only for mode=continue/test
        
        # only for testing
        'n_eval_episodes': [None,'rs', 'int'],   # required only for mode=test 
        'render': [False,'o', 'bool'],           # option to plot during test via self.render() 
        }
        
        #---------------------------------
        #A2C Card
        #---------------------------------
        self.a2c_dict={
        'mode': ['train', 'r', 'str'],  # either train, continue, or test
        'time_steps': [50000,'rs','int'], # required only for mode=continue/test 
        'flag': [False, 'o', 'bool'],   # this will be activated if READ A2C is found in input
        'casename': ['a2c', 'o', 'str'],           # prefix for logging
        'ncores': [1,'o','int'],                   # multiple cores are available
        'n_steps': [5,'o','int'],                  # see manual
        'gamma': [0.99, 'o', 'float'],             # see manual
        'learning_rate': [0.0007, 'o', 'float'],   # see manual
        'vf_coef': [0.25, 'o', 'float'],           # see manual
        'max_grad_norm': [0.5, 'o', 'float'],      # see manual
        'ent_coef': [0.01, 'o', 'float'],          # see manual
        'momentum': [0.0, 'o', 'float'],          # see manual
        'alpha': [0.99, 'o', 'float'],             # see manual
        'epsilon': [1e-5, 'o', 'float'],           # see manual
        'lr_schedule': ['constant', 'o', 'str'],   # see manual
        
        'model_load_path': [None,'rs', 'str'], # required only for mode=continue/test
        
        # for plotting
        'check_freq': [1000,'o', 'int'],        # see manual
        'avg_episodes': [20,'o', 'int'],        # see manual
        'tensorboard': [False,'o','bool'],      # tensorboard view option, see manual for usage
        
        # only for testing
        'n_eval_episodes': [None,'rs', 'int'], # required only for mode=test 
        'render': [False,'o', 'bool'],         # option to plot during test via self.render()         
        }

        #---------------------------------
        #ACER Card
        #---------------------------------
        self.acer_dict={
        'mode': ['train', 'r', 'str'],  # either train, continue, or test
        'time_steps': [50000,'rs','int'], # required only for mode=continue/test 
        'flag': [False, 'o', 'bool'],   # this will be activated if READ A2C is found in input
        'casename': ['acer', 'o', 'str'],           # prefix for logging
        'ncores': [1,'o','int'],                   # multiple cores are available
        'n_steps': [20,'o','int'],                  # see manual
        'gamma': [0.99, 'o', 'float'],             # see manual
        'learning_rate': [0.0007, 'o', 'float'],   # see manual
        'q_coef': [0.5, 'o', 'float'],           # see manual
        'max_grad_norm': [10, 'o', 'float'],      # see manual
        'ent_coef': [0.01, 'o', 'float'],          # see manual
        'alpha': [0.99, 'o', 'float'],             # see manual
        'lr_schedule': ['linear', 'o', 'str'],   # see manual
        'rprop_alpha': [0.99, 'o', 'float'],           # see manual
        'rprop_epsilon': [1e-5, 'o', 'float'],           # see manual
        'buffer_size': [5000, 'o', 'int'],           # see manual
        'replay_ratio': [4, 'o', 'float'],           # see manual
        'replay_start': [1000, 'o', 'int'],           # see manual
        'correction_term': [10, 'o', 'float'],           # see manual
        'trust_region': [True, 'o', 'bool'],           # see manual
        'delta': [1, 'o', 'float'],           # see manual
        
        'model_load_path': [None,'rs', 'str'], # required only for mode=continue/test
        
        # for plotting
        'check_freq': [1000,'o', 'int'],        # see manual
        'avg_episodes': [20,'o', 'int'],        # see manual
        'tensorboard': [False,'o','bool'],      # tensorboard view option, see manual for usage
        
        # only for testing
        'n_eval_episodes': [None,'rs', 'int'], # required only for mode=test 
        'render': [False,'o', 'bool'],         # option to plot during test via self.render()         
        }
        
        #---------------------------------
        # GA card
        #---------------------------------
        self.ga_dict={
        'flag': [False, 'o', 'bool'],     # this will be activated if READ GA is found in input
        'mode': [None, 'r', 'str'],       # either assign (casmo) or shuffle (simulate3)
        'pop': [None, 'r', 'int'],        # number of population surviving from lambda per generation
        'lambda': [None, 'o', 'int'],     # number of total population created per generation
        'ngen': [None, 'r', 'int'],       # total number of generations 
        'casename': ['ga', 'o', 'str'],   # prefix for logging/results
        'lbound': [None, 'o', 'vec'],     # see manual for definition
        'ubound': [None, 'o', 'vec'],     # see manual for definition
        'type': [None, 'o', 'strvec'],     # see manual for definition
        'kbs_path': [None, 'o', 'str'],  # path to use kbs
        'kbs_frac': [0.1, 'o', 'float'],  # path to use kbs
        'ncores': [1,'o','int'],          # only 1 core
        'indpb': [0.05,'o','float'],      # independent prob for input attribute swap
        'cxpb': [0.5, 'o', 'float'],      # crossover prob
        'mutpb': [0.2, 'o', 'float'],      # mutation prob
        'smin': [0.01, 'o', 'float'],      # minimum strategy value in strategy vector
        'smax': [0.5, 'o', 'float'],      # maximum strategy value in strategy vector
        'check_freq': [1, 'o', 'int']    # update master log every check_freq (min is 1, max is ngen)
        }
        
        #---------------------------------
        # SA card
        #---------------------------------
        self.sa_dict={
        'flag': [False, 'o', 'bool'],     # this will be activated if READ SA is found in input
        'steps': [None, 'r', 'int'],      # total number of annealing steps (env calls)
        'casename': ['sa', 'o', 'str'],   # prefix for logging/results
        'ncores': [1,'o','int'],          # only 1 core
        'cooling': ['fast','o','str'],
        'swap': ['kbs', 'r', 'str'], # either singleswap, dualswap, quadswap, or fullswap, kbs
        'indpb_kbs': [0.5, 'o', 'float'], # prob to use kbs
        'kbs_path': [None, 'rs', 'str'], # path to kbs dataset
        'indpb': [0.05,'o','float'],      # independent prob for input attribute swap
        'lbound': [None, 'o', 'vec'],     # see manual for definition
        'ubound': [None, 'o', 'vec'],     # see manual for definition
        'type': [None, 'o', 'strvec'],     # see manual for definition
        'Tmin': [1, 'o', 'int'],       # Min T to end the annealing process 
        'Tmax': [25000,'o','int'],         # Max T to start the annealing process 
        'initstate': [None,'o','vec'],    # First state, random sample by default
        'check_freq': [50, 'o', 'int'],    # update master log every check_freq
        'avg_step': [50, 'o', 'int']      # update master log every check_freq (min is 1, max is steps)
        }



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
"""
Created on Fri Jun 19 17:23:10 2020

@author: Majdi
"""

import random 
import numpy as np

#alpha0=0, anneal_alpha=False, anneal_steps=None, alpha_end=1
class ExperienceReplay:
    def __init__(self, size):
        #"""
        #:param size (int): the max size of the memory
        #"""
        random.seed(1)
        self.size=size
        self.storage = []
        self.next_indx=0    
        
    @property
    def memory(self):
        #"""
        #content of the memory buffer
        #[(np.ndarray, float, str)]: 
        #"""
        return self.storage

    @property
    def buffer_size(self):
        #"""Max memory capacity"""
        return self.size
    

    def memory_can_sample(self, n_samples):
        #"""
        #Check if n_samples samples can be sampled from the memory.
        #:param n_samples (int):  number of samples to draw
        #:return (bool): whether we can sample or not
        #"""
        return len(self.storage) >= n_samples

    def memory_is_full(self):
        #"""
        #Check whether the replay buffer is full or not.
        #:return: (bool)
        #"""
        return len(self.storage) == self.buffer_size

    def add(self, xvec, obj, method=None):
        #"""
        #add a new sample to the memory
        #:param xvec (list): a sample of input attributes (x1,x2,...,xn)
        #:param obj (float): objective value of xvec
        #:param method (string): method of which this sample belongs to
        #"""

        
        #check if multiple or single samples is to be added 
        if type(obj) is list: # multiple samples 
            if method:
                data = [(x, o, m) for x,o,m in zip(xvec,obj,method)]
            else:
                data = [(x, o) for x,o in zip(xvec,obj)]
            
            for sample in data:
                # check if sample is in memory
                if sample not in self.storage:
                    if self.next_indx >= len(self.storage):
                        self.storage.append(sample)
                    else:
                        self.storage[self.next_indx] = sample
                    self.next_indx = (self.next_indx + 1) % self.size
            
        else: #single sample
            if method:
                data = (xvec, obj, method)
            else:
                data = (xvec, obj)
            
            # check if sample is in memory
            if data not in self.storage:
                if self.next_indx >= len(self.storage):
                    self.storage.append(data)
                else:
                    self.storage[self.next_indx] = data
                self.next_indx = (self.next_indx + 1) % self.size
                    
    def calc_priorities(self, alpha):
        #"""
        #calculate priorties for each memory sample
        #:param alpha: priortization value 
        #:return
        #  - list of normalized priorities 
        #"""
        #Fixed :)
        
        self.storage.sort(key=lambda e: e[1], reverse=True)
        ranks=np.array(range(1,len(self.storage)+1))
        ranks=1/ranks
        priors=ranks**alpha/np.sum(ranks**alpha)
        assert np.round(np.sum(priors)) == 1.0, 'the calculated priorties are not normalized'
        return priors

    def sample(self, batch_size=1, mode='uniform', alpha=0.5, seed=1):
        #"""
        #Sample a batch of experiences.
        #:param batch_size: number of samples to draw from the memory
        #:param mode (str): the priortization mode
        #  -"uniform": uniform sampling
        #  -"greedy": always use the best samples
        #  -"prior": priortized replay with alpha value
        #:param alpha: the prioritization ceoffcient between 0 (no priority) and 1 (full priority)
        #:return:
        #    - batch_size of samples in a list of tuples [(np.ndarray, float, str),...,(np.ndarray, float, str)]
        #"""
        if mode=='uniform': # uniform sampling
            idxs = [random.randint(0, len(self.storage) - 1) for _ in range(batch_size)]
            return [self.storage[i] for i in idxs]
        elif mode=='greedy': #greedy mode (always take the highest)
            #Fixed :) 
            self.storage.sort(key=lambda e: e[1], reverse=True)
            return self.storage[:batch_size]
        elif mode=='prior':  #priortized replay
            priors=self.calc_priorities(alpha=alpha)
            np.random.seed(seed)
            idxs = np.random.choice(range(len(self.storage)),p=priors, size=batch_size)
            return [self.storage[i] for i in idxs]
        else:
            raise ('unknown mode is entered for experience replay: either uniform, greedy, or prior are allowed')
    
    def remove_duplicates(self):
        
        #TODO: Time consuming step!!!
        if len(self.storage[0])==3: #(x, obj, method) tuple
            seen = []
            # using list comprehension 
            self.storage = [(a, b, c) for a, b, c in self.storage 
                     if not (a in seen or seen.append(a))]
        elif len(self.storage[0])==2:   #(x, obj) tuple
            seen = []
            # using list comprehension 
            self.storage = [(a, b) for a, b in self.storage 
                     if not (a in seen or seen.append(a))]
        else:
            raise ('memory content is corrupted and cannot be filtered')

#if __name__=='__main__':
#    random.seed(1)
#    per=ExperienceReplay(size=100, mode='uniform')
#    
#    for i in range(100):
#        sample=(np.ones((4))*random.randint(1,10), random.randint(1,10), random.choice(['ga','sa','pso']))
#        per.add(xvec=sample[0], obj=sample[1], method=sample[2])
#    
#    x=per.storage
#    samples=per.sample(10,alpha=1)
#    print([item[1] for item in samples])
  
    


import numpy as np


def unique(sorted_array):
    """
    More efficient implementation of np.unique for sorted arrays
    :param sorted_array: (np.ndarray)
    :return:(np.ndarray) sorted_array without duplicate elements
    """
    if len(sorted_array) == 1:
        return sorted_array
    left = sorted_array[:-1]
    right = sorted_array[1:]
    uniques = np.append(right != left, True)
    return sorted_array[uniques]


class SegmentTree(object):
    def __init__(self, capacity, operation, neutral_element):
        """
        Build a Segment Tree data structure.

        https://en.wikipedia.org/wiki/Segment_tree

        Can be used as regular array that supports Index arrays, but with two
        important differences:

            a) setting item's value is slightly slower.
               It is O(lg capacity) instead of O(1).
            b) user has access to an efficient ( O(log segment size) )
               `reduce` operation which reduces `operation` over
               a contiguous subsequence of items in the array.

        :param capacity: (int) Total size of the array - must be a power of two.
        :param operation: (lambda (Any, Any): Any) operation for combining elements (eg. sum, max) must form a
            mathematical group together with the set of possible values for array elements (i.e. be associative)
        :param neutral_element: (Any) neutral element for the operation above. eg. float('-inf') for max and 0 for sum.
        """
        assert capacity > 0 and capacity & (capacity - 1) == 0, "capacity must be positive and a power of 2."
        self._capacity = capacity
        self._value = [neutral_element for _ in range(2 * capacity)]
        self._operation = operation
        self.neutral_element = neutral_element

    def _reduce_helper(self, start, end, node, node_start, node_end):
        if start == node_start and end == node_end:
            return self._value[node]
        mid = (node_start + node_end) // 2
        if end <= mid:
            return self._reduce_helper(start, end, 2 * node, node_start, mid)
        else:
            if mid + 1 <= start:
                return self._reduce_helper(start, end, 2 * node + 1, mid + 1, node_end)
            else:
                return self._operation(
                    self._reduce_helper(start, mid, 2 * node, node_start, mid),
                    self._reduce_helper(mid + 1, end, 2 * node + 1, mid + 1, node_end)
                )

    def reduce(self, start=0, end=None):
        """
        Returns result of applying `self.operation`
        to a contiguous subsequence of the array.

            self.operation(arr[start], operation(arr[start+1], operation(... arr[end])))

        :param start: (int) beginning of the subsequence
        :param end: (int) end of the subsequences
        :return: (Any) result of reducing self.operation over the specified range of array elements.
        """
        if end is None:
            end = self._capacity
        if end < 0:
            end += self._capacity
        end -= 1
        return self._reduce_helper(start, end, 1, 0, self._capacity - 1)

    def __setitem__(self, idx, val):
        # indexes of the leaf
        idxs = idx + self._capacity
        self._value[idxs] = val
        if isinstance(idxs, int):
            idxs = np.array([idxs])
        # go up one level in the tree and remove duplicate indexes
        idxs = unique(idxs // 2)
        while len(idxs) > 1 or idxs[0] > 0:
            # as long as there are non-zero indexes, update the corresponding values
            self._value[idxs] = self._operation(
                self._value[2 * idxs],
                self._value[2 * idxs + 1]
            )
            # go up one level in the tree and remove duplicate indexes
            idxs = unique(idxs // 2)

    def __getitem__(self, idx):
        assert np.max(idx) < self._capacity
        assert 0 <= np.min(idx)
        return self._value[self._capacity + idx]


class SumSegmentTree(SegmentTree):
    def __init__(self, capacity):
        super(SumSegmentTree, self).__init__(
            capacity=capacity,
            operation=np.add,
            neutral_element=0.0
        )
        self._value = np.array(self._value)

    def sum(self, start=0, end=None):
        """
        Returns arr[start] + ... + arr[end]

        :param start: (int) start position of the reduction (must be >= 0)
        :param end: (int) end position of the reduction (must be < len(arr), can be None for len(arr) - 1)
        :return: (Any) reduction of SumSegmentTree
        """
        return super(SumSegmentTree, self).reduce(start, end)

    def find_prefixsum_idx(self, prefixsum):
        """
        Find the highest index `i` in the array such that
            sum(arr[0] + arr[1] + ... + arr[i - i]) <= prefixsum for each entry in prefixsum

        if array values are probabilities, this function
        allows to sample indexes according to the discrete
        probability efficiently.

        :param prefixsum: (np.ndarray) float upper bounds on the sum of array prefix
        :return: (np.ndarray) highest indexes satisfying the prefixsum constraint
        """
        if isinstance(prefixsum, float):
            prefixsum = np.array([prefixsum])
        assert 0 <= np.min(prefixsum)
        assert np.max(prefixsum) <= self.sum() + 1e-5
        assert isinstance(prefixsum[0], float)

        idx = np.ones(len(prefixsum), dtype=int)
        cont = np.ones(len(prefixsum), dtype=bool)

        while np.any(cont):  # while not all nodes are leafs
            idx[cont] = 2 * idx[cont]
            prefixsum_new = np.where(self._value[idx] <= prefixsum, prefixsum - self._value[idx], prefixsum)
            # prepare update of prefixsum for all right children
            idx = np.where(np.logical_or(self._value[idx] > prefixsum, np.logical_not(cont)), idx, idx + 1)
            # Select child node for non-leaf nodes
            prefixsum = prefixsum_new
            # update prefixsum
            cont = idx < self._capacity
            # collect leafs
        return idx - self._capacity


class MinSegmentTree(SegmentTree):
    def __init__(self, capacity):
        super(MinSegmentTree, self).__init__(
            capacity=capacity,
            operation=np.minimum,
            neutral_element=float('inf')
        )
        self._value = np.array(self._value)

    def min(self, start=0, end=None):
        """
        Returns min(arr[start], ...,  arr[end])

        :param start: (int) start position of the reduction (must be >= 0)
        :param end: (int) end position of the reduction (must be < len(arr), can be None for len(arr) - 1)
        :return: (Any) reduction of MinSegmentTree
        """
        return super(MinSegmentTree, self).reduce(start, end)



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- encoding: utf-8 -*-
#'''
#@File    :   FNEAT Class based on RNEAT 
#@Time    :   2021/07/27 10:17:37
#@Author  :   Xubo GU 
#@Email   :   guxubo@alumni.sjtu.edu.cn
#'''

import numpy as np  
import neat        
import pickle
import random
import os
from multiprocessing import Pool
from neorl.rl.make_env import CreateEnvironment
from neorl.utils.seeding import set_neorl_seed

class FNEAT(object):
    """
    Feedforward NeuroEvolution of Augmenting Topologies (FNEAT)
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization (RL is default to ``max``)       
    :param fit: (function) the fitness function
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param config: (dict) dictionary of RNEAT hyperparameters, see **Notes** below for available hyperparameters to change
    :param ncores: (int) number of parallel processors
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, fit, bounds, config, ncores=1, seed=None):
        
        set_neorl_seed(seed)
        
        self.ncores=ncores
        self.mode=mode
        self.bounds = bounds
        self.fit=fit
        self.nx=len(self.bounds)
        default_config = self.basic_config()  #construct the default config file
        self.config = self.modify_config(default_config, config) #modify the config based on user input
        #force the required NEAT variables (Do not change)
        self.config['NEAT']['fitness_criterion'] = "max"
        self.config['DefaultGenome']['num_inputs'] = self.nx
        self.config['DefaultGenome']['num_outputs'] = self.nx
        self.episode_length=self.config['NEAT']['pop_size']
        
        self.env=CreateEnvironment(method='fneat', fit=self.fit, ncores=1, 
                      bounds=self.bounds, mode=self.mode, episode_length=self.episode_length)
        
    def eval_genomes(self, genomes, config):

        for genome_id, genome in genomes:
            
            if self.x0: # input user's data 
                ob=self.x0.copy()
            else: # no user's data 
                ob = self.env.reset()
                
            net = neat.nn.FeedForwardNetwork.create(genome, config)
            local_fit = float("-inf")
            counter = 0
            xpos = 0
            done = False
            
            while not done:
                
                nnOutput = net.activate(ob)
                ob, rew, done, info = self.env.step(nnOutput)
                xpos = info['x']  
                                
                if rew > local_fit:
                    local_fit = rew
                    counter = 0
                else:
                    counter += 1
                
                if rew > self.best_fit:
                    self.best_fit=rew
                    self.best_x=xpos.copy()

                #--mir
                if self.mode=='max':
                    self.best_fit_correct=self.best_fit
                    local_fit_correct=local_fit
                else:
                    self.best_fit_correct=-self.best_fit
                    local_fit_correct=-local_fit
                
                if done or counter == self.episode_length:
                    done = True
                    self.history['global_fitness'].append(self.best_fit_correct)
                    self.history['local_fitness'].append(local_fit_correct)        
                    #print('best fit:', self.best_fit_correct)
                    
                genome.fitness = rew

    def genome_worker(self, genome, config):
        # parallel worker that passes different eval_genomes to diffrent cores
        worker = NEATWorker(genome, config, self.episode_length, self.x0, env=self.env)
        fitness, local_fit, xpos=worker.work()
        return fitness, local_fit, xpos
             
    def evolute(self, ngen, x0=None, save_best_net=False, checkpoint_itv=None, startpoint=None, verbose=False):
        """
        This function evolutes the FNEAT algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list) initial position of the NEAT (must have same size as the ``x`` variable)
        :param save_best_net: (bool) save the winner neural network to a pickle file
        :param checkpoint_itv: (int) generation frequency to save checkpoints for restarting purposes (e.g. 1: save every generation, 10: save every 10 generations)
        :param startpoint: (str) name/path to the checkpoint file to use to start the search (the checkpoint file can be saved by invoking the argument ``checkpoint_itv``) 
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        
        self.history={'global_fitness': [], 'local_fitness':[]}
        self.best_fit=float("-inf")
        self.verbose=verbose        
        
        self.x0=x0
        if self.x0 is not None:
            self.x0 = list(self.x0)
            assert len(self.x0) == self.nx, '--error: the length of x0 ({}) MUST equal the size of the bounds variable ({})'.format(len(self.x0), self.nx)
        
        # transfer dict-type config to neat type
        path = os.path.dirname(__file__)
        file_data = []
        for section, content in self.config.items():
            file_data.append('\n')
            file_data.append('[' + section + ']'+'\n')
            for key, val in content.items():
                file_data.append(key + '=' + str(val) + '\n')

        tmp_file = os.path.join(path, 'tmp_config')
        if not os.path.exists(tmp_file):
            file = open(tmp_file, 'w')
            file.close()
            print('--debug: Temporary config file created ... ')

        with open(tmp_file, 'w') as ft:
            for line in file_data:
                ft.write(line)
            print('--debug: Modified config has been written ... ') 

        config = neat.Config(neat.DefaultGenome, neat.DefaultReproduction,
                             neat.DefaultSpeciesSet, neat.DefaultStagnation,
                             tmp_file)
        
        os.remove(tmp_file)
        print('--debug: Temporary config file removed ...')

        p = neat.Population(config)

        # train from start or checkpoint
        if startpoint:
            print('\nTrain model from {}'.format(startpoint))
            p = neat.Checkpointer.restore_checkpoint(startpoint)
            
            if verbose:
                p.add_reporter(neat.StdOutReporter(True))
                stats = neat.StatisticsReporter() 
                p.add_reporter(stats)
            if checkpoint_itv:
                p.add_reporter(neat.Checkpointer(checkpoint_itv))

            #winner = p.run(self.eval_genomes, ngen) # total gen = startpoint+ngen

        else: 
            print('\n--debug: Train model from the start.')
            if verbose:
                p.add_reporter(neat.StdOutReporter(True))
                stats = neat.StatisticsReporter()
                p.add_reporter(stats)
            if checkpoint_itv: # checkpoint saving interval
                print('\n--debug: Save model every {} epochs'.format(checkpoint_itv))
                p.add_reporter(neat.Checkpointer(checkpoint_itv))
            #winner = p.run(self.eval_genomes, ngen)
        
        #parallel runner
        if self.ncores > 1:
            print('--debug: FNEAT is running in parallel with {} cores ...'.format(self.ncores))
            pe = ParallelEvaluator(self.ncores, self.genome_worker, self.mode)
            winner = p.run(pe.evaluate, ngen)
            self.best_fit_correct=pe.best_fit_correct
            self.best_x=pe.best_x
            self.history=pe.history
        else:
            winner = p.run(self.eval_genomes, ngen)
            
        
        if save_best_net:
            with open('winner-net', 'wb') as output:
                pickle.dump(winner, output, 1)
                print('--debug: Winner net saved ...')
            
        if verbose:
            print('------------------------ FNEAT Summary --------------------------')
            print('Best fitness (y) found:', self.best_fit_correct)
            print('Best individual (x) found:', self.best_x)
            print('--------------------------------------------------------------')
            
        return self.best_x, self.best_fit_correct, self.history

    def modify_config(self, config_basic, config_modify_dict):
        #"""
        #config_basic: NEAT's basic configs, type: dict
        #config_modify_dict: storing configs to be changed, type: dict
        #"""
        
        para_change_list = [p for p in config_modify_dict.keys()] # parameters changed 
        
        for section, paras in config_basic.items():
            print('--debug: Dealing with section [{}] ...'.format(str(section)))
            for key, value in paras.items():
                if key in para_change_list:
                    # idx = para_change_list.index(key)
                    config_basic[section][key] = config_modify_dict[key]
                    print('--debug: Change parameter "{}" to "{}" successfully!'\
                        .format(str(key),str(config_modify_dict[key])))
    
        print('--debug: ************NEAT config file is constructed!************')
        return config_basic

    def basic_config(self):
        # This function builds the default NEAT config file.
        a = {
            'NEAT':{
            'fitness_criterion':'max',  #mir: default is max
            'fitness_threshold': 1e5,
            'pop_size': 30,
            'reset_on_extinction': True,
            'no_fitness_termination': False
            },
    
            'DefaultGenome':{
            'activation_default':'identity',
            'activation_mutate_rate':0.05,
            'activation_options': 'sigmoid',
    
            'aggregation_default': 'random',
            'aggregation_mutate_rate':0.05,
            'aggregation_options': 'sum product min max mean median maxabs',
            
            'single_structural_mutation': False,
            'structural_mutation_surer': 'default',
            
            'bias_init_type': 'gaussian',
            'bias_init_mean': 0.05,
            'bias_init_stdev': 1.0,
            'bias_max_value': 30.0,
            'bias_min_value':  -30.0,
            'bias_mutate_power': 0.5,
            'bias_mutate_rate': 0.7,
            'bias_replace_rate': 0.1,
    
            'compatibility_disjoint_coefficient': 1.0,
            'compatibility_weight_coefficient': 0.5,
    
            'conn_add_prob': 0.5,
            'conn_delete_prob': 0.1,
    
            'enabled_default': True,
            'enabled_mutate_rate': 0.2,
            'enabled_rate_to_true_add': 0.0,
            'enabled_rate_to_false_add': 0.0,
    
            'feed_forward': True,
            'initial_connection': 'partial_nodirect 0.5',
    
            'node_add_prob': 0.5,
            'node_delete_prob': 0.5,
    
            'num_hidden': 1,
            'num_inputs': None,  #mir: must be defined by the developer
            'num_outputs': None,  #mir: must be defined by the developer
            
            'response_init_type': 'gaussian',
            'response_init_mean': 1.0,
            'response_init_stdev': 0.05,
            'response_max_value': 30.0,
            'response_min_value': -30.0,
            'response_mutate_power': 0.1,
            'response_mutate_rate': 0.75,
            'response_replace_rate': 0.1,
            
            'weight_init_type': 'gaussian', 
            'weight_init_mean': 0.1,
            'weight_init_stdev': 1.0,
            'weight_max_value': 30,
            'weight_min_value': -30,
            'weight_mutate_power': 0.5,   
            'weight_mutate_rate': 0.8,
            'weight_replace_rate': 0.1
            },
    
            'DefaultSpeciesSet':{
            'compatibility_threshold': 2.5
            },
    
            'DefaultStagnation':{
            'species_fitness_func': 'max',
            'max_stagnation': 50,
            'species_elitism': 0
            },
    
            'DefaultReproduction':{
            'elitism': 1,
            'survival_threshold': 0.3,
            'min_species_size': 2
            }
        }
    
        return a

class NEATWorker(object):
    #This class provides a path to allow passing 
    #different workers with different genomes and configs
    #Inputs:
        #genome: genome structure
        #config: config for the genome
        #episode_length: epsiode length for logging if defined in the original FNEAT class
        #x0: initial guess if defined in the original FNEAT class
        #env: enviroment class
    
    def __init__(self, genome, config, episode_length, x0, env):
        self.genome = genome
        self.config = config
        self.episode_length=episode_length
        self.x0=x0
        self.env=env
        
    def work(self):
        
        if self.x0: # input user's data 
            ob=self.x0.copy()
        else: # no user's data 
            ob = self.env.reset()
            
        net = neat.nn.FeedForwardNetwork.create(self.genome, self.config)
        local_fit = float("-inf")
        counter = 0
        xpos = 0
        done = False
        
        while not done:
            
            nnOutput = net.activate(ob)
            ob, rew, done, info = self.env.step(nnOutput)
            xpos = info['x']  
                            
            if rew > local_fit:
                local_fit = rew
                counter = 0
            else:
                counter += 1

        if done or counter == self.episode_length:
            done = True
                
        return rew, local_fit, xpos
    


class ParallelEvaluator(object):
    def __init__(self, num_workers, eval_function, mode, timeout=None):
        #"""
        #Runs evaluation functions in parallel subprocesses
        #in order to evaluate multiple genomes at once.
        
        #eval_function should take one argument, a tuple of
        #(genome object, config object), and return
        #a single float (the genome's fitness).
        #"""
        self.num_workers = num_workers
        self.eval_function = eval_function
        self.timeout = timeout
        self.pool = Pool(num_workers)
        self.history={'global_fitness': [], 'local_fitness':[]}
        self.best_fit=float("-inf")
        self.mode=mode
        
    def __del__(self):
        self.pool.close() # should this be terminate?
        self.pool.join()

    def evaluate(self, genomes, config):
        jobs = []
        for ignored_genome_id, genome in genomes:
            jobs.append(self.pool.apply_async(self.eval_function, (genome, config)))

        # assign the fitness back to each genome
        for job, (ignored_genome_id, genome) in zip(jobs, genomes):
            genome.fitness, local_fit, xpos = job.get(timeout=self.timeout)
        
            if genome.fitness > self.best_fit:
                self.best_fit=genome.fitness
                self.best_x=xpos.copy()
    
            #--mir
            if self.mode=='max':
                self.best_fit_correct=self.best_fit
                local_fit_correct=local_fit
            else:
                self.best_fit_correct=-self.best_fit
                local_fit_correct=-local_fit

            self.history['global_fitness'].append(self.best_fit_correct)
            self.history['local_fitness'].append(local_fit_correct)